import numpy as np
from torch import nn
import torch
import re
import math
import functools
from sklearn.metrics import average_precision_score


def get_ap_score(y_true, y_scores):
    """
    Get average precision score between 2 1-d numpy arrays

    Args:
        y_true: batch of true labels
        y_scores: batch of confidence scores

    Returns:
        sum of batch average precision
    """
    scores = average_precision_score(y_true=y_true, y_score=y_scores, average='samples')
    return scores


def val_mae(pred_counts, gt_counts):
    n = len(gt_counts)
    true_count = np.ones(n) * (-1)
    pred_count = np.ones(n) * (-1)

    for index in range(n):
        true_count[index] = gt_counts[index]
        pred_count[index] = pred_counts[index]
        # mae = (np.abs(true_count[:index+1] - pred_count[:index+1])).mean()
    score_dict = {}
    assert not np.any(true_count == (-1))
    assert not np.any(pred_count == (-1))
    score_dict["MAE"] = (np.abs(true_count - pred_count)).mean()
    print(f"score_dict: {score_dict}")
    return score_dict


def mae(pred_counts, gt_counts):
    assert len(pred_counts) == len(gt_counts)
    n = len(gt_counts)
    absolute_e_list = [abs(b_i - a_i) for a_i, b_i in zip(pred_counts, gt_counts)]
    sum_e_list = sum(absolute_e_list)
    mae = sum_e_list / n
    return mae


def rmse(pred_counts, gt_counts):
    assert len(pred_counts) == len(gt_counts)
    n = len(gt_counts)

    absolute_e_list = [abs(b_i - a_i) * abs(b_i - a_i) for a_i, b_i in zip(pred_counts, gt_counts)]
    sum_e_list = sum(absolute_e_list)
    sum_e_list = sum_e_list / n
    rmse = math.sqrt(sum_e_list)
    return rmse


def mape(pred_counts, gt_counts):
    assert len(pred_counts) == len(gt_counts)
    n = len(gt_counts)

    absolute_e_list = [abs(b_i - a_i) / b_i for a_i, b_i in zip(pred_counts, gt_counts)]
    sum_e_list = sum(absolute_e_list)
    sum_e_list = sum_e_list / n
    mape = 100 * sum_e_list
    return mape


def _take_channels(*xs, ignore_channels=None):
    if ignore_channels is None:
        return xs
    else:
        channels = [channel for channel in range(xs[0].shape[1])
                    if channel not in ignore_channels]
        xs = [torch.index_select(x, dim=1, index=torch.tensor(channels))
              for x in xs]
        return xs


def _threshold(x, threshold=None):
    if threshold is not None:
        return (x > threshold).type(x.dtype)
    else:
        return x


def iou_npy(pr, gt, eps=1e-7, threshold=None, ignore_channels=None):
    """Calculate Intersection over Union between ground truth and prediction
    Args:
        pr (torch.Tensor): predicted tensor
        gt (torch.Tensor):  ground truth tensor
        eps (float): epsilon to avoid zero division
        threshold: threshold for outputs binarization
    Returns:
        float: IoU (Jaccard) score
    """

    intersection = np.sum(gt * pr)
    print(intersection)
    union = np.sum(gt) + np.sum(pr) - intersection + eps
    print(union)
    return (intersection + eps) / union


def iou(pr, gt, eps=1e-7, threshold=None, ignore_channels=None):
    """Calculate Intersection over Union between ground truth and prediction
    Args:
        pr (torch.Tensor): predicted tensor
        gt (torch.Tensor):  ground truth tensor
        eps (float): epsilon to avoid zero division
        threshold: threshold for outputs binarization
    Returns:
        float: IoU (Jaccard) score
    """

    pr = _threshold(pr, threshold=threshold)
    pr, gt = _take_channels(pr, gt, ignore_channels=ignore_channels)

    intersection = torch.sum(gt * pr)
    union = torch.sum(gt) + torch.sum(pr) - intersection + eps
    return (intersection + eps) / union


jaccard = iou


def f_score(pr, gt, beta=1, eps=1e-7, threshold=None, ignore_channels=None):
    """Calculate F-score between ground truth and prediction
    Args:
        pr (torch.Tensor): predicted tensor
        gt (torch.Tensor):  ground truth tensor
        beta (float): positive constant
        eps (float): epsilon to avoid zero division
        threshold: threshold for outputs binarization
    Returns:
        float: F score
    """

    pr = _threshold(pr, threshold=threshold)
    pr, gt = _take_channels(pr, gt, ignore_channels=ignore_channels)

    tp = torch.sum(gt * pr)
    fp = torch.sum(pr) - tp
    fn = torch.sum(gt) - tp

    score = (((1 + beta ** 2) * tp + eps) /
             ((1 + beta ** 2) * tp + beta ** 2 * fn + fp + eps))

    return score


def accuracy(pr, gt, threshold=0.5, ignore_channels=None):
    """Calculate accuracy score between ground truth and prediction
    Args:
        pr (torch.Tensor): predicted tensor
        gt (torch.Tensor):  ground truth tensor
        eps (float): epsilon to avoid zero division
        threshold: threshold for outputs binarization
    Returns:
        float: precision score
    """
    pr = _threshold(pr, threshold=threshold)
    pr, gt = _take_channels(pr, gt, ignore_channels=ignore_channels)

    tp = torch.sum(gt == pr)
    score = tp / gt.view(-1).shape[0]
    return score


def precision(pr, gt, eps=1e-7, threshold=None, ignore_channels=None):
    """Calculate precision score between ground truth and prediction
    Args:
        pr (torch.Tensor): predicted tensor
        gt (torch.Tensor):  ground truth tensor
        eps (float): epsilon to avoid zero division
        threshold: threshold for outputs binarization
    Returns:
        float: precision score
    """

    pr = _threshold(pr, threshold=threshold)
    pr, gt = _take_channels(pr, gt, ignore_channels=ignore_channels)

    tp = torch.sum(gt * pr)
    fp = torch.sum(pr) - tp

    score = (tp + eps) / (tp + fp + eps)

    return score


def recall(pr, gt, eps=1e-7, threshold=None, ignore_channels=None):
    """Calculate Recall between ground truth and prediction
    Args:
        pr (torch.Tensor): A list of predicted elements
        gt (torch.Tensor):  A list of elements that are to be predicted
        eps (float): epsilon to avoid zero division
        threshold: threshold for outputs binarization
    Returns:
        float: recall score
    """

    pr = _threshold(pr, threshold=threshold)
    pr, gt = _take_channels(pr, gt, ignore_channels=ignore_channels)

    tp = torch.sum(gt * pr)
    fn = torch.sum(gt) - tp

    score = (tp + eps) / (tp + fn + eps)

    return score


class BaseObject(nn.Module):

    def __init__(self, name=None):
        super().__init__()
        self._name = name

    @property
    def __name__(self):
        if self._name is None:
            name = self.__class__.__name__
            s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
            return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()
        else:
            return self._name


class Activation(nn.Module):
    def __init__(self, activation):
        super().__init__()
        if activation is None or activation == 'identity':
            self.activation = nn.Identity()
        elif activation == 'sigmoid':
            self.activation = torch.sigmoid
        elif activation == 'softmax2d':
            self.activation = functools.partial(torch.softmax, dim=1)
        elif callable(activation):
            self.activation = activation
        else:
            raise ValueError

    def forward(self, x):
        return self.activation(x)


def _fast_hist(label_pred, label_true, num_classes):
    mask = (label_true >= 0) & (label_true < num_classes)
    hist = np.bincount(
        num_classes * label_true[mask].astype(int) +
        label_pred[mask], minlength=num_classes ** 2).reshape(num_classes, num_classes)
    return hist


def calc_mAP(predictions, gts, num_classes=2):
    hist = np.zeros((num_classes, num_classes))
    for lp, lt in zip(predictions, gts):
        hist += _fast_hist(lp.flatten(), lt.flatten(), num_classes)
    # axis 0: gt, axis 1: prediction
    acc = np.diag(hist).sum() / hist.sum()
    acc_cls = np.diag(hist) / hist.sum(axis=1)
    acc_cls = np.nanmean(acc_cls)
    iu = np.diag(hist) / (hist.sum(axis=1) +
                          hist.sum(axis=0) - np.diag(hist))
    # iu = np.diag(hist) / (hist.sum(axis=1) +
    #                       - np.diag(hist))
    mean_iu = np.nanmean(iu)
    freq = hist.sum(axis=1) / hist.sum()
    fwavacc = (freq[freq > 0] * iu[freq > 0]).sum()
    return acc, acc_cls, mean_iu, fwavacc


class Metric(BaseObject):
    pass


class IoU(Metric):
    __name__ = 'iou_score'

    def __init__(self, eps=1e-7, threshold=0.5, activation=None, ignore_channels=None, **kwargs):
        super().__init__(**kwargs)
        self.eps = eps
        self.threshold = threshold
        self.activation = Activation(activation)
        self.ignore_channels = ignore_channels

    def forward(self, y_pr, y_gt):
        y_pr = self.activation(y_pr)
        return iou(
            y_pr, y_gt,
            eps=self.eps,
            threshold=self.threshold,
            ignore_channels=self.ignore_channels,
        )


class Fscore(Metric):

    def __init__(self, beta=1, eps=1e-7, threshold=0.5, activation=None, ignore_channels=None, **kwargs):
        super().__init__(**kwargs)
        self.eps = eps
        self.beta = beta
        self.threshold = threshold
        self.activation = Activation(activation)
        self.ignore_channels = ignore_channels

    def forward(self, y_pr, y_gt):
        y_pr = self.activation(y_pr)
        return f_score(
            y_pr, y_gt,
            eps=self.eps,
            beta=self.beta,
            threshold=self.threshold,
            ignore_channels=self.ignore_channels,
        )


class Accuracy(Metric):

    def __init__(self, threshold=0.5, activation=None, ignore_channels=None, **kwargs):
        super().__init__(**kwargs)
        self.threshold = threshold
        self.activation = Activation(activation)
        self.ignore_channels = ignore_channels

    def forward(self, y_pr, y_gt):
        y_pr = self.activation(y_pr)
        return accuracy(
            y_pr, y_gt,
            threshold=self.threshold,
            ignore_channels=self.ignore_channels,
        )


class Recall(Metric):

    def __init__(self, eps=1e-7, threshold=0.5, activation=None, ignore_channels=None, **kwargs):
        super().__init__(**kwargs)
        self.eps = eps
        self.threshold = threshold
        self.activation = Activation(activation)
        self.ignore_channels = ignore_channels

    def forward(self, y_pr, y_gt):
        y_pr = self.activation(y_pr)
        return recall(
            y_pr, y_gt,
            eps=self.eps,
            threshold=self.threshold,
            ignore_channels=self.ignore_channels,
        )


class Precision(Metric):

    def __init__(self, eps=1e-7, threshold=0.5, activation=None, ignore_channels=None, **kwargs):
        super().__init__(**kwargs)
        self.eps = eps
        self.threshold = threshold
        self.activation = Activation(activation)
        self.ignore_channels = ignore_channels

    def forward(self, y_pr, y_gt):
        y_pr = self.activation(y_pr)
        return precision(
            y_pr, y_gt,
            eps=self.eps,
            threshold=self.threshold,
            ignore_channels=self.ignore_channels,
        )


def compute_jaccard(preds_masks_all, targets_masks_all, num_classes=21):

    tps = np.zeros((num_classes, ))
    fps = np.zeros((num_classes, ))
    fns = np.zeros((num_classes, ))
    # counts = np.zeros((num_classes, ))

    for mask_pred, mask_gt in zip(preds_masks_all, targets_masks_all):
        # print(mask_pred.shape)
        # print(mask_gt.shape)
        # print(mask_pred.size())
        bs, h, w = mask_pred.size()
        assert bs == mask_gt.size(0), "Batch size mismatch"
        assert h == mask_gt.size(1), "Width mismatch"
        assert w == mask_gt.size(2), "Height mismatch"

        mask_pred = mask_pred.view(bs, 1, -1)
        mask_gt = mask_gt.view(bs, 1, -1)

        # ignore ambiguous
        # mask_pred[mask_gt == 255] = 255

        for label in range(num_classes):
            mask_pred_ = (mask_pred == label).float()
            mask_gt_ = (mask_gt == label).float()

            tps[label] += (mask_pred_ * mask_gt_).sum().item()
            diff = mask_pred_ - mask_gt_
            fps[label] += np.maximum(0., diff).float().sum().item()
            fns[label] += np.maximum(0., -diff).float().sum().item()

    eps = 1e-3
    jaccards  = [None] * num_classes
    precision = [None] * num_classes
    recall    = [None] * num_classes
    for i in range(num_classes):
        tp = tps[i]
        fn = fns[i]
        fp = fps[i]
        jaccards[i]  = tp / max(eps, fn + fp + tp)
        precision[i] = tp / max(eps, tp + fp)
        recall[i]    = tp / max(eps, tp + fn)

    return jaccards, precision, recall



import torch
import torch.nn as nn
import torch.nn.functional as F

import torch.nn.parallel
import numpy as np


class BlurPool(nn.Module):
    def __init__(self, channels, pad_type='reflect', filt_size=4, stride=2, pad_off=0):
        super(BlurPool, self).__init__()
        self.filt_size = filt_size
        self.pad_off = pad_off
        self.pad_sizes = [int(1. * (filt_size - 1) / 2), int(np.ceil(1. * (filt_size - 1) / 2)), int(1. * (filt_size - 1) / 2), int(np.ceil(1. * (filt_size - 1) / 2))]
        self.pad_sizes = [pad_size + pad_off for pad_size in self.pad_sizes]
        self.stride = stride
        self.off = int((self.stride - 1) / 2.)
        self.channels = channels

        if (self.filt_size == 1):
            a = np.array([1., ])
        elif (self.filt_size == 2):
            a = np.array([1., 1.])
        elif (self.filt_size == 3):
            a = np.array([1., 2., 1.])
        elif (self.filt_size == 4):
            a = np.array([1., 3., 3., 1.])
        elif (self.filt_size == 5):
            a = np.array([1., 4., 6., 4., 1.])
        elif (self.filt_size == 6):
            a = np.array([1., 5., 10., 10., 5., 1.])
        elif (self.filt_size == 7):
            a = np.array([1., 6., 15., 20., 15., 6., 1.])
        else:
            raise ValueError(self.filt_size)

        filt = torch.Tensor(a[:, None] * a[None, :])
        filt = filt / torch.sum(filt)
        self.register_buffer('filt', filt[None, None, :, :].repeat((self.channels, 1, 1, 1)))

        self.pad = get_pad_layer(pad_type)(self.pad_sizes)

    def forward(self, inp):
        if (self.filt_size == 1):
            if (self.pad_off == 0):
                return inp[:, :, ::self.stride, ::self.stride]
            else:
                return self.pad(inp)[:, :, ::self.stride, ::self.stride]
        else:
            return F.conv2d(self.pad(inp), self.filt, stride=self.stride, groups=inp.shape[1])


def get_pad_layer(pad_type):
    if (pad_type in ['refl', 'reflect']):
        PadLayer = nn.ReflectionPad2d
    elif (pad_type in ['repl', 'replicate']):
        PadLayer = nn.ReplicationPad2d
    elif (pad_type == 'zero'):
        PadLayer = nn.ZeroPad2d
    else:
        print('Pad type [%s] not recognized' % pad_type)
    return PadLayer


class double_conv(nn.Module):
    def __init__(self, in_ch, out_ch):
        super(double_conv, self).__init__()
        self.conv = nn.Sequential(
            nn.Conv2d(in_ch, out_ch, 3, padding=1),
            nn.BatchNorm2d(out_ch),
            nn.ReLU(inplace=True),
            nn.Conv2d(out_ch, out_ch, 3, padding=1),
            nn.BatchNorm2d(out_ch),
            nn.ReLU(inplace=True)
        )

    def forward(self, x):
        x = self.conv(x)
        return x


class inconv(nn.Module):
    def __init__(self, in_ch, out_ch):
        super(inconv, self).__init__()
        self.conv = double_conv(in_ch, out_ch)

    def forward(self, x):
        x = self.conv(x)
        return x


class down(nn.Module):
    def __init__(self, in_ch, out_ch):
        super(down, self).__init__()
        self.mpconv = nn.Sequential(
            nn.MaxPool2d(2, stride=1),
            BlurPool(in_ch, filt_size=4, stride=2),
            double_conv(in_ch, out_ch)
        )

    def forward(self, x):
        x = self.mpconv(x)
        return x


class up(nn.Module):
    def __init__(self, in_ch, out_ch):
        super(up, self).__init__()
        self.up = nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True)

        self.conv = double_conv(in_ch, out_ch)

    def forward(self, x1, x2):
        #x1 = self.up(x1)
        x1 = nn.functional.interpolate(x1, scale_factor=2, mode='nearest')

        # input is CHW
        diffY = x2.size()[2] - x1.size()[2]
        diffX = x2.size()[3] - x1.size()[3]

        x1 = F.pad(x1, (diffX // 2, diffX - diffX // 2, diffY // 2, diffY - diffY // 2))

        x = torch.cat([x2, x1], dim=1)
        x = self.conv(x)
        return x


class outconv(nn.Module):
    def __init__(self, in_ch, out_ch):
        super(outconv, self).__init__()
        self.conv = nn.Conv2d(in_ch, out_ch, 1, bias=False)

    def forward(self, x):
        x = self.conv(x)
        return x


# U-Net
class UNet(nn.Module):
    def __init__(self, in_channels, out_channels):
        super(UNet, self).__init__()

        self.inc = inconv(in_channels, 64)
        self.down1 = down(64, 128)
        self.down2 = down(128, 256)
        self.down3 = down(256, 512)
        self.down4 = down(512, 512)
        self.up1 = up(1024, 256)
        self.up2 = up(512, 128)
        self.up3 = up(256, 64)
        self.up4 = up(128, 64)
        self.outc = outconv(64, out_channels)

    def forward(self, x):
        x1 = self.inc(x)
        x2 = self.down1(x1)
        x3 = self.down2(x2)
        x4 = self.down3(x3)
        x5 = self.down4(x4)

        x = self.up1(x5, x4)
        x = self.up2(x, x3)
        x = self.up3(x, x2)
        x = self.up4(x, x1)
        x = self.outc(x)
        return x



def fix_gelu_issue(method):
    # Torch 1.12 added an approximate parameter that our old models dont
    # have. Monkey patch it in.
    # https://github.com/pytorch/pytorch/pull/61439
    for name, mod in method.named_modules():
        if mod.__class__.__name__ == 'GELU':
            if not hasattr(mod, 'approximate'):
                mod.approximate = 'none'


def fix_package_modules():
    # Monkey Patch torch.package
    import sys
    if sys.version_info[0:2] >= (3, 10):
        try:
            from torch.package import _stdlib
            _stdlib._get_stdlib_modules = lambda: sys.stdlib_module_names
        except Exception:
            pass



#!/usr/bin/env python3
# PYTHON_ARGCOMPLETE_OK
import sys
import ubelt as ub

__devnotes__ = """
# We may want to delay actual imports, gdal import time can be excessive

python -X importtime -c "import watch"
WATCH_HACK_IMPORT_ORDER="" python  -X importtime -m watch.cli find_dvc
"""


def main(cmdline=True, **kw):
    """
    The watch command line interface
    """
    modnames = [
        'watch_coco_stats',
        'geojson_site_stats',
        'torch_model_stats',
        'coco_spectra',

        # 'geotiffs_to_kwcoco'  # TODO: cleanup and add

        'coco_visualize_videos',
        'gifify',

        'find_dvc',

        'coco_add_watch_fields',
        'coco_align',
        'coco_clean_geotiffs',
        'coco_average_features',
        'coco_time_combine',

        'reproject_annotations',

        'kwcoco_to_geojson',
        'run_metrics_framework',

        'crop_sites_to_regions',
        'coco_remove_bad_images',
        'coco_average_features',

        # 'mlops_cli',
    ]

    cmd_alias = {
        'watch.cli.torch_model_stats': ['model_stats', 'model_info'],
        'watch.cli.geojson_site_stats': ['site_stats'],
        'watch.cli.watch_coco_stats': ['stats'],
        'watch.cli.coco_visualize_videos': ['visualize'],
        'watch.cli.coco_align': ['align', 'coco_align_geotiffs'],
        'watch.cli.reproject_annotations': ['reproject', 'project'],
        'watch.cli.coco_add_watch_fields': ['add_fields'],
        'watch.cli.coco_spectra': ['spectra', 'intensity_histograms'],
        'watch.cli.run_metrics_framework': ['iarpa_eval'],
        'watch.cli.coco_clean_geotiffs': ['clean_geotiffs'],
        'watch.cli.kwcoco_to_geojson': ['run_tracker'],
        'watch.cli.find_dvc': ['dvcdir'],
        'watch.cli.gifify': ['animate'],
        'watch.cli.coco_average_features': ['average_features', 'ensemble'],
        'watch.cli.coco_time_combine': ['time_combine'],
        'watch.cli.crop_sites_to_regions': ['crop_sitemodels'],
        'watch.cli.coco_remove_bad_images': ['remove_bad_images'],
        'watch.cli.mlops_cli': ['mlops'],
    }

    module_lut = {}
    for name in modnames:
        mod = ub.import_module_from_name('watch.cli.{}'.format(name))
        module_lut[name] = mod

    module_lut['schedule'] = ub.import_module_from_name('watch.mlops.schedule_evaluation')
    module_lut['manager'] = ub.import_module_from_name('watch.mlops.manager')
    module_lut['aggregate'] = ub.import_module_from_name('watch.mlops.aggregate')

    # Create a list of all submodules with CLI interfaces
    cli_modules = list(module_lut.values())

    import os
    from scriptconfig.modal import ModalCLI
    import watch
    WATCH_LOOSE_CLI = os.environ.get('WATCH_LOOSE_CLI', '')

    # https://emojiterra.com/time/
    fun_header = ub.codeblock(
        '''
        🛰️ ⌚🛰️           🧠       🛰️ ⌚🛰️
        🌍🌎🌏        👁️   👁️      🌏🌎🌍
        ''')

    boring_description =  ub.codeblock(
        f'''
        🌐🌐🌐 The GEO-WATCH CLI 🌐🌐🌐

        An open source research and production environment for image and video
        segmentation and detection with geospatial awareness.

        Developed by [link=https://www.kitware.com/]Kitware[/link]. Funded by the [link=https://www.iarpa.gov/research-programs/smart]IARPA SMART[/link] challenge.

        Version: {watch.__version__}
        ''')

    FUN = (os.getenv('FUN', '') or 1) and not os.getenv('NOFUN', '')
    if FUN:
        description = fun_header + '\n' + boring_description
    else:
        description = boring_description

    modal = ModalCLI(description)

    modal.__class__.version = watch.__version__

    for cli_module in cli_modules:

        cli_subconfig = None
        if not hasattr(cli_module, '__config__'):
            if hasattr(cli_module, 'modal'):
                continue
            raise AssertionError('We are only supporting scriptconfig CLIs')
        # scriptconfig cli pattern
        cli_subconfig = cli_module.__config__

        if not hasattr(cli_subconfig, 'main'):
            if hasattr(cli_module, 'main'):
                main_func = cli_module.main
                # Hack the main function into the config
                cli_subconfig.main = main_func
            else:
                raise AssertionError(f'No main function for {cli_module}')

        # Update configs to have aliases / commands attributes
        cli_modname = cli_module.__name__
        cli_rel_modname = cli_modname.split('.')[-1]

        cmdname_aliases = ub.oset()
        alias = getattr(cli_subconfig, '__alias__', [])
        if isinstance(alias, str):
            alias = [alias]
        command = getattr(cli_module, '__command__', None)
        command = getattr(cli_subconfig, '__command__', command)
        if command is not None:
            cmdname_aliases.add(command)
        cmdname_aliases.update(alias)
        cmdname_aliases.update(cmd_alias.get(cli_modname, []) )
        cmdname_aliases.add(cli_rel_modname)
        # parserkw = {}
        primary_cmdname = cmdname_aliases[0]
        secondary_cmdnames = cmdname_aliases[1:]
        if not isinstance(primary_cmdname, str):
            raise AssertionError(primary_cmdname)
        cli_subconfig.__command__ = primary_cmdname
        cli_subconfig.__alias__ = list(secondary_cmdnames)
        modal.register(cli_subconfig)

    ret = modal.run(strict=not WATCH_LOOSE_CLI)
    return ret


if __name__ == '__main__':
    """
    CommandLine:
        python -m watch.cli --help
        python -m watch.cli.watch_coco_stats
    """
    sys.exit(main())



# flake8: noqa
import kwcoco
import ndsampler
import ubelt as ub
import numpy as np
import matplotlib.pyplot as plt
import torch
from sklearn.cluster import KMeans
from sklearn.manifold import TSNE
from watch.tasks.rutgers_material_seg.datasets.iarpa_dataset import SequenceDataset

visualize_images = False
coco_fpath = ub.expandpath('/media/native/data/data/smart_watch_dvc/drop0_aligned_msi/material_labels.kwcoco.json')
dset = kwcoco.CocoDataset(coco_fpath)

sampler = ndsampler.CocoSampler(dset)

# # print(sampler)
number_of_timestamps, h, w = 5, 128, 128
window_dims = (number_of_timestamps, h, w)  # [t,h,w]
input_dims = (h, w)

# # channels = 'r|g|b|gray|wv1'
# channels = 'r|g|b'
channels = 'red|green|blue|nir|swir16|swir22'
# channels = 'nir'``
dataset = SequenceDataset(sampler, window_dims, input_dims, channels)
loader = dataset.make_loader(batch_size=1)

k = 80
kmeans = KMeans(n_clusters=k, random_state=0)
kmeans_tsne = KMeans(n_clusters=k, random_state=0)

for batch_index, batch in enumerate(loader):
    # pdb.set_trace()
    image_data = batch['inputs']['im'].data[0]  # [b,c,t,h,w]
    print(f"image_data min: {image_data.min()}, image_data max: {image_data.max()}")
    b, c, t, h, w = image_data.shape
    mask_data = batch['label']['class_masks'].data[0]  # len(mask_data) = b
    mask_data = torch.stack(mask_data)

    image_show = np.array(image_data).transpose(0, 2, 3, 4, 1)  # /50000 # visualize 0 indexed in batch
    image_show = image_show / image_show.max()
    print(f"image min: {image_show.min()}, image max: {image_show.max()}")
    # print(image_show.shape)
    # image_show = image_show[0,0,:,:,:]
    # plt.imshow(image_show[0,0,:,:,:3])
    # plt.show()
    # mask_show = np.array(mask_data) # [b,t,h,w]

    image_data = image_data.view(b, c * t, h * w)
    # print(image_data.shape)
    image_data = torch.transpose(image_data, 1, 2)
    # print(image_data.shape)
    image_data = torch.flatten(image_data, start_dim=0, end_dim=1)
    # print(image_data.shape)
    # image_data = torch.transpose(image_data,0,1)
    # print(image_data.shape)
    out_feat_embed = TSNE(n_components=2).fit_transform(image_data)
    kmeans_tsne.fit(out_feat_embed)
    # data = image_data
    # data = out_feat_embed
    kmeans.fit(image_data)
    cluster_centers = kmeans.cluster_centers_
    cluster_labels = kmeans.labels_
    # y_kmeans = kmeans.predict(data)
    y_kmeans_tse = kmeans_tsne.predict(out_feat_embed)
    # print(cluster_centers)
    # print(cluster_labels)
    prediction = cluster_labels.reshape(h, w)
    prediction_no_bg = np.ma.masked_where(prediction == 0, prediction)
    # print(f"image_data: {image_data.shape}, mask: {mask_data.shape}")
    # print(f"image min: {image_show.min()}, image max: {image_show.max()}")
    plt.scatter(out_feat_embed[:, 0], out_feat_embed[:, 1], c=y_kmeans_tse, marker='.', cmap='tab20c')
    plt.scatter(kmeans_tsne.cluster_centers_[:, 0], kmeans_tsne.cluster_centers_[:, 1], c='black', s=200, alpha=0.5)
    plt.show()

    figure = plt.figure(figsize=(15, 15))
    ax1 = figure.add_subplot(1, 5, 1)
    ax2 = figure.add_subplot(1, 5, 2)
    ax3 = figure.add_subplot(1, 5, 3)
    ax4 = figure.add_subplot(1, 5, 4)
    ax5 = figure.add_subplot(1, 5, 5)
    # ax6 = figure.add_subplot(2,4,2)
    # ax7 = figure.add_subplot(2,4,1)
    # ax8 = figure.add_subplot(2,4,2)

    ax1.imshow(image_show[0, 0, :, :, :3])
    ax2.imshow(image_show[0, 1, :, :, :3])
    ax3.imshow(image_show[0, 2, :, :, :3])
    ax4.imshow(image_show[0, 3, :, :, :3])
    # ax1.imshow(image_show[0,0,:,:,:])
    ax5.imshow(prediction, vmin=0, vmax=k, cmap='Set1', interpolation='nearest')
    # ax3.imshow(image_show[0,0,:,:,:])
    # ax3.imshow(prediction_no_bg, alpha=0.6, cmap='Set1')
#     ax4.scatter(x_clusters_scatters, y_clusters_scatters, color=(len(x_clusters_scatters)//channels)*['red','green','blue','yellow','black'])

    plt.show()

    # if visualize_images:
    #     mask_show = mask_show[0] # [b,t,h,w]
    #     image_show = image_show[0]
    #     figure = plt.figure(figsize=(10,10))
    #     axes = {}
    #     for i in range(1,2*t+1):
    #         axes[i] = figure.add_subplot(2,t,i)
    #     for key in axes.keys():
    #         if key <= t:
    #             axes[key].imshow(image_show[key-1,:,:,:])
    #         else:
    #             axes[key].imshow(mask_show[key-t-1,:,:],vmin=-1, vmax=7)
    #     figure.tight_layout()
    #     plt.show()



import ubelt as ub
import os
from watch.utils import util_path


def tryget_dvc_dpath():
    _default = ub.expandpath('$HOME/data/dvc-repos/smart_watch_dvc')
    dvc_dpath = os.environ.get('DVC_DPATH', _default)
    dvc_dpath = util_path.coercepath(dvc_dpath)
    if not dvc_dpath.exists():
        import pytest
        pytest.skip('this test depends on data in DVC_DPATH')
    return dvc_dpath


def test_load_uky_models():
    import torch
    dvc_dpath = tryget_dvc_dpath()
    uky_dpath = (dvc_dpath / 'models/uky_invariants/sort_augment_overlap')
    checkpoint_fpaths = list(uky_dpath.glob('*.ckpt'))
    states = []
    for fpath in checkpoint_fpaths:
        loaded = torch.load(fpath)
        states.append(loaded)

    for state in states:
        info = ub.dict_isect(state, {'epoch', 'hyper_parameters', 'global_step'})
        print('info = {}'.format(ub.urepr(info, nl=2)))
        print(state['epoch'])
        print(state['hyper_parameters'])
        print(state['hparams_name'])
        print(state['global_step'])

    # inv_sort, inv_overlap, inv_shared, or inv_augment?

    r"""
        /models/uky_invariants/sort_augment_overlap/S2_drop1-S2-L8-aligned-old.0.ckpt
        /models/uky_invariants/sort_augment_overlap/L8_drop1-S2-L8-aligned-old.0.ckpt
        /models/uky_invariants/sort_augment_overlap/LS_drop1-S2-L8-aligned-old.0.ckpt
        /models/uky_invariants/sort_augment_overlap/S2_drop1-S2-aligned-old.0.ckpt


    Notes on prediction:
        DVC_DPATH=/home/joncrall/data/dvc-repos/smart_watch_dvc
        CHECKPOINT_FPATH=$DVC_DPATH/models/uky_invariants/sort_augment_overlap/S2_drop1-S2-L8-aligned-old.0.ckpt
        CHECKPOINT_FPATH=$DVC_DPATH/models/uky_invariants/sort_augment_overlap/L8_drop1-S2-L8-aligned-old.0.ckpt
        CHECKPOINT_FPATH=$DVC_DPATH/models/uky_invariants/sort_augment_overlap/LS_drop1-S2-L8-aligned-old.0.ckpt
        CHECKPOINT_FPATH=$DVC_DPATH/models/uky_invariants/sort_augment_overlap/S2_drop1-S2-aligned-old.0.ckpt

        python -m watch stats $DVC_DPATH/drop1-S2-L8-aligned/data.kwcoco.json

        python -m watch.tasks.invariants.predict \
            --sensor S2 \
            --input_kwcoco $DVC_DPATH/drop1-S2-L8-aligned/data.kwcoco.json \
            --output_kwcoco $DVC_DPATH/drop1-S2-L8-aligned/_partial_uky_pred_S2.kwcoco.json \
            --ckpt_path $DVC_DPATH/models/uky_invariants/sort_augment_overlap/S2_drop1-S2-L8-aligned-old.0.ckpt

        python -m watch.tasks.invariants.predict \
            --sensor L8 \
            --input_kwcoco $DVC_DPATH/drop1-S2-L8-aligned/data.kwcoco.json \
            --output_kwcoco $DVC_DPATH/drop1-S2-L8-aligned/_partial_uky_pred_L8.kwcoco.json \
            --ckpt_path $DVC_DPATH/models/uky_invariants/sort_augment_overlap/L8_drop1-S2-L8-aligned-old.0.ckpt

        cd $DVC_DPATH/drop1-S2-L8-aligned
        kwcoco union --src _partial_uky_pred_S2.kwcoco.json _partial_uky_pred_L8.kwcoco.json --dst uky_invariants.kwcoco.json

        python ~/code/watch/watch/cli/coco_combine_features.py --src \
                _partial_uky_pred_S2.kwcoco.json \
                _partial_uky_pred_L8.kwcoco.json \
                --dst uky_invariants.kwcoco.json

        python -m watch stats _partial_uky_pred_S2.kwcoco.json
        python -m watch stats _partial_uky_pred_L8.kwcoco.json
        python -m watch stats uky_invariants.kwcoco.json

        cd $DVC_DPATH/drop1-S2-L8-aligned
        cd $DVC_DPATH/drop1-S2-L8-aligned
        python ~/code/watch/watch/cli/coco_combine_features.py --src \
                data.kwcoco.json \
                landcover.kwcoco.json \
                uky_invariants.kwcoco.json \
                --dst ./combo_data.kwcoco.json

        rice_field|cropland|water|inland_water|river_or_stream|sebkha|snow_or_ice_field|bare_ground|sand_dune|built_up|grassland|brush|forest|wetland|road

        python ~/data/dvc-repos/smart_watch_dvc/dev/coco_show_auxiliary.py combo_data.kwcoco.json --channel2 bare_ground
        python ~/data/dvc-repos/smart_watch_dvc/dev/coco_show_auxiliary.py combo_data.kwcoco.json --channel2 water
        python ~/data/dvc-repos/smart_watch_dvc/dev/coco_show_auxiliary.py combo_data.kwcoco.json --channel2 forest
        python ~/data/dvc-repos/smart_watch_dvc/dev/coco_show_auxiliary.py combo_data.kwcoco.json --channel2 inland_water
        python ~/data/dvc-repos/smart_watch_dvc/dev/coco_show_auxiliary.py combo_data.kwcoco.json --channel2 river_or_stream
    """



"""
This is a very task-specific file containing logic to parse fusion pipeline
metrics for BAS and SC.


Generate:
    import liberator
    lib = liberator.Liberator()
    import networkx as nx
    lib.add_dynamic(nx.DiGraph)
    print(ub.highlight_code(lib.current_sourcecode()))
"""

import numpy as np
import warnings
import json
import ubelt as ub
import io
import xdev
import pandas as pd
import re
from watch.utils import util_time
from watch.utils import util_pattern


@ub.memoize
@xdev.profile
def parse_json_header(fpath):
    """
    Ideally the information we need is in the first few bytes of the json file
    """
    from watch.utils import ijson_ext
    import zipfile
    if zipfile.is_zipfile(fpath):
        # We have a compressed json file, but we can still read the header
        # fairly quickly.
        zfile = zipfile.ZipFile(fpath)
        names = zfile.namelist()
        assert len(names) == 1
        member = names[0]
        # Stream the header directly from the zipfile.
        file = zfile.open(member, 'r')
    else:
        # Normal json file
        file = open(fpath, 'r')

    with file:
        # import ijson
        # We only expect there to be one info section
        # try:
        #     # Try our extension if the main library fails (due to NaN)
        #     info_section_iter = ijson.items(file, prefix='info')
        #     info_section = next(info_section_iter)
        # except ijson.IncompleteJSONError:
        # Try our extension if the main library fails (due to NaN)
        # file.seek(0)

        # Nans are too frequent, only use our extension
        info_section_iter = ijson_ext.items(file, prefix='info')
        info_section = next(info_section_iter)
    return info_section


@xdev.profile
def trace_json_lineage(fpath):
    """
    We will expect a json file to contain a top-level "info" section that
    indicates how it is derived.

    fpath = '/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/models/fusion/Aligned-Drop4-2022-08-08-TA1-S2-L8-ACC/pred/trk/Drop4_BAS_Retrain_V002_epoch=31-step=16384.pt.pt/Aligned-Drop4-2022-08-08-TA1-S2-L8-ACC_data.kwcoco/trk_pxl_b788335d/trk_poly_f2218f0b/tracks.kwcoco.json'
    """

    info_section = parse_json_header(fpath)

    # TODO:
    # uniqify by uuid
    for proc in list(find_info_items(info_section, {'process', 'process_context'})):
        name = proc['properties']['name']
        if name not in {'coco_align_geotiffs', 'coco_align'}:
            print(f'name={name}')
            print(proc['properties']['start_timestamp'])
            print(proc['properties']['emissions']['run_id'])
            print('proc = {}'.format(ub.urepr(proc, nl=2)))
        # print(proc['properties']['name'])


# def trace_kwcoco_lineage(fpath):
@xdev.profile
def load_iarpa_evaluation(fpath):
    """
    Args:
        fpath (PathLike | str):
            path to the IARPA summary json file

    Returns:
        Dict: containing keys:
            metrics -
                which just contains a flat Dict[str, float] metric dictionary
            iarpa_info -
                which contains ALL of the information parsed out of the summary json file.

    Ignore:
        fpath = '/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/models/fusion/Drop4-BAS/eval/trk/package_epoch0_step41.pt.pt/Drop4-BAS_KR_R001.kwcoco/trk_pxl_fd9e1a95/trk_poly_9f08fb8c/merged/summary2.json'
    """
    iarpa_json = _load_json(fpath)
    metrics = {}
    unique_regions = set()
    if 'best_bas_rows' in iarpa_json:
        best_bas_rows = pd.read_json(
            io.StringIO(json.dumps(iarpa_json['best_bas_rows'])),
            orient='table')
        bas_row = best_bas_rows.loc['__macro__'].reset_index().iloc[0]

        metrics.update({
            'bas_tp': bas_row['tp sites'],
            'bas_fp': bas_row['fp sites'],
            'bas_fn': bas_row['fn sites'],
            'bas_ntrue': bas_row['total sites'],
            'bas_npred': bas_row['proposed slices'],
            'bas_ppv': bas_row['precision'],
            'bas_tpr': bas_row['recall (PD)'],
            'bas_ffpa': bas_row['ffpa'],
            'bas_f1': bas_row['F1'],
            'rho': bas_row['rho'],
            'tau': bas_row['tau'],
            'bas_space_FAR': bas_row['spatial FAR'],
            'bas_time_FAR': bas_row['temporal FAR'],
            'bas_image_FAR': bas_row['images FAR'],
        })
        alpha = 1.0
        metrics['bas_faa_f1'] = metrics['bas_f1'] * (1 - metrics['bas_ffpa']) ** alpha

        unique_regions.update(
            best_bas_rows.index.levels[best_bas_rows.index.names.index('region_id')].difference({'__macro__', '__micro__'})
        )

    if 'sc_df' in iarpa_json:
        sc_json_data = iarpa_json['sc_df']
        sc_json_text = json.dumps(sc_json_data)
        try:
            sc_df = pd.read_json(io.StringIO(sc_json_text), orient='table')
        except pd.errors.IntCastingNaNError:
            # This seems like a pandas bug. It looks like it can save a Int64
            # with NaN exteions, but it can't load it back in.
            sc_json_data = iarpa_json['sc_df']
            walker = ub.IndexableWalker(sc_json_data)
            for path, val in walker:
                if path[-1] == 'extDtype' and val == 'Int64':
                    walker[path] = 'number'
                if path[-1] == 'type' and val == 'integer':
                    walker[path] = 'number'
            sc_json_text = json.dumps(sc_json_data)
            sc_df = pd.read_json(io.StringIO(sc_json_text), orient='table')

        metrics.update({
            # 'mean_f1': sc_df.loc['F1'].mean(),
            'sc_macro_f1': sc_df.loc['__macro__']['F1'].mean(),
            'macro_f1_siteprep': sc_df.loc['__macro__', 'Site Preparation']['F1'],
            'macro_f1_active': sc_df.loc['__macro__', 'Site Preparation']['F1'],
        })

        if '__micro__' in sc_df.index:
            # Not sure why micro sometimes is not included.
            metrics['sc_micro_f1'] = sc_df.loc['__micro__']['F1'].mean()
        else:
            metrics['sc_micro_f1'] = np.nan

        unique_regions.update(
            sc_df.index.levels[sc_df.index.names.index('region_id')].difference({'__macro__', '__micro__'})
        )

    # quick and dirty way to get access to single-region results
    region_ids = ','.join(sorted(unique_regions))
    iarpa_json['region_ids'] = region_ids

    iarpa_result = {
        'metrics': metrics,
        'iarpa_json': iarpa_json,
    }
    return iarpa_result


@xdev.profile
def load_bas_poly_eval(fpath, expt_dvc_dpath=None, arg_prefix='trk.'):
    """
    fpath = ub.Path('/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_testpipe/eval/flat/bas_poly_eval/bas_poly_eval_id_1ad531cc/poly_eval.json')
    expt_dvc_dpath = None
    arg_prefix = ''
    """
    iarpa_result = load_iarpa_evaluation(fpath)
    iarpa_json = iarpa_result['iarpa_json']
    metrics = iarpa_result['metrics']

    tracker_info = iarpa_json['parent_info']

    if 0:
        from watch.utils.util_dotdict import indexable_to_graph
        data = tracker_info
        graph = indexable_to_graph(data)
        for n in graph.nodes:
            name = n.split('.')[-1]
            if name in {'args', 'config', 'machine', 'disk_info', 'device_info'}:
                graph.nodes[n]['collapse'] = True
        from cmd_queue.util.util_networkx import write_network_text
        import rich
        write_network_text(graph, path=rich.print, end='', max_depth=None)

    # param_types = parse_tracker_params(tracker_info, expt_dvc_dpath, arg_prefix=arg_prefix)

    # extra_attrs = _add_prefix(arg_prefix + 'poly.metrics.', metrics)
    info = {
        'fpath': fpath,
        'metrics': metrics,
        # 'param_types': param_types,
        # 'other': {
        #     'extra_attrs': extra_attrs,
        # },
        'json_info': iarpa_json,
    }
    return info


@xdev.profile
def load_iarpa_poly_eval(fpath):
    metrics, iarpa_info = load_iarpa_evaluation(fpath)
    info = {
        'fpath': fpath,
        'metrics': metrics,
        'json_info': iarpa_info,
    }
    return info


@xdev.profile
def load_sc_poly_eval(fpath, expt_dvc_dpath=None, arg_prefix='act.'):
    metrics, iarpa_info = load_iarpa_evaluation(fpath)

    # tracker_info = iarpa_info.get('parent_info', None)
    # if tracker_info is not None:
    #     param_types = parse_tracker_params(tracker_info, expt_dvc_dpath, arg_prefix=arg_prefix)
    # else:
    #     param_types = {}
    # # Hack to grab information that we should have already had.
    # HACK_HANDLE_CROPPED_AND_TRACK_PARAMS = 1
    # if HACK_HANDLE_CROPPED_AND_TRACK_PARAMS:
    #     try:
    #         trk_param_types, extra_attrs = _handle_crop_and_trk_params(param_types, expt_dvc_dpath)
    #     except Exception:
    #         trk_param_types = {}
    #         extra_attrs = {}
    #     param_types.update(trk_param_types)
    # else:
    #     extra_attrs = {}
    # extra_attrs.update(_add_prefix('act.poly.metrics.', metrics))

    info = {
        'fpath': fpath,
        'metrics': metrics,
        # 'param_types': param_types,
        # 'other': {
        #     'extra_attrs': extra_attrs,
        # },
        'json_info': iarpa_info,
    }
    return info


@xdev.profile
def _handle_crop_and_trk_params(param_types, expt_dvc_dpath):
    from watch.mlops import expt_manager
    act_pxl_test_dset = param_types['act.pxl']['act.pxl.test_dataset']
    state = expt_manager.ExperimentState('*', '*')

    dset_source = None
    try:
        crop_attrs = ub.udict(state._parse_pattern_attrs(
            state.templates['crop_fpath'], act_pxl_test_dset))
    except RuntimeError:
        # Probably because we used the truth
        dset_source = 'truth'
    else:
        crop_attrs = {}
        dset_source = 'tracker'

    crop_dataset = _load_json(act_pxl_test_dset)
    crop_item = list(find_info_items(
        crop_dataset['info'],
        {'process', 'process_context'},
        {'coco_align_geotiffs', 'coco_align'}
    ))[-1]
    # This is the path to either truth or the tracks we cropped from
    region_fpath = crop_item['properties']['args']['regions']

    if dset_source == 'tracker':
        # Our input was a bas tracking output
        try:
            trk_attrs = ub.udict(state._parse_pattern_attrs(
                state.templates['pred_trk_poly_site_summaries_fpath'],
                region_fpath))
        except Exception:
            trk_attrs = ub.udict(state._parse_pattern_attrs(
                state.templates['pred_trk_poly_sites_fpath'],
                region_fpath))
    else:
        trk_attrs = {}

    extra_attrs = ub.udict(crop_attrs) | ub.udict(trk_attrs)

    if dset_source == 'tracker':
        trk_poly_data = _load_json(region_fpath)
        trk_poly_info = trk_poly_data['info']
        trk_param_types = parse_tracker_params(trk_poly_info, expt_dvc_dpath, arg_prefix='trk.')
    else:
        trk_param_types = {}
    return trk_param_types, extra_attrs


@xdev.profile
def parse_tracker_params(tracker_info, expt_dvc_dpath=None, arg_prefix=''):
    """
    Args:
        tracker_info (List[Dict]):
            This should be the "info" section of a tracker result (i.e. "info"
            in the kwcoco json or geojson manifest file), which is a list of
            process context dictionaries. One of these dictionaries will have a
            process name "watch.cli.kwcoco_to_geojson" or
            "watch.cli.run_tracker", and that item will contain the config used
            to run the tracker. It may also contain an "extra.pred_info"
            property containing the pixel prediction params, and that may
            contain the training configuration.

    Note:
        This is tricky because we need to find a way to differentiate if this
        was a trk or bas tracker.
    """
    track_item = find_track_item(tracker_info)

    track_item = _handle_process_item(track_item)

    if 0:
        from watch.utils.util_dotdict import indexable_to_graph
        graph = indexable_to_graph(track_item)
        for n in graph.nodes:
            name = n.split('.')[-1]
            if name in {
                    # 'args',
                    # 'config',
                    'machine', 'disk_info', 'device_info', 'fit_config', 'emissions'}:
                graph.nodes[n]['collapse'] = True
        from cmd_queue.util.util_networkx import write_network_text
        import rich
        write_network_text(graph, path=rich.print, end='', max_depth=None)

    if 'extra' in track_item['properties']:
        pred_info = track_item['properties']['extra']['pred_info']
    else:
        raise AssertionError

    param_types = parse_pred_pxl_params(
        pred_info, expt_dvc_dpath, arg_prefix=arg_prefix)

    poly_resources = parse_resource_item(track_item, arg_prefix=(arg_prefix + 'poly.'))
    param_types[arg_prefix + 'poly.resource'] = poly_resources

    track_args = track_item['properties']['config']
    track_config = relevant_track_config(track_args, arg_prefix=arg_prefix)
    param_types[arg_prefix + 'poly'] = track_config
    return param_types


@xdev.profile
def _handle_process_item(item):
    """
    Json data written by the process context has changed over time slightly.
    Consolidate different usages until a consistent API and usage patterns are
    established.

    """
    assert item['type'] in {'process', 'process_context'}
    props = item['properties']

    needs_modify = 0

    config = props.get('config', None)
    args = props.get('args', None)
    if config is None:
        # Use args if config is not available
        config = args
        needs_modify = True

    FIX_BROKEN_SCRIPTCONFIG_HANDLING = 1
    if FIX_BROKEN_SCRIPTCONFIG_HANDLING:
        if '_data' in config:
            config = config['_data']
            needs_modify = True
        if '_data' in args:
            args = args['_data']
            needs_modify = True

    assert 'pred_info' not in item, 'should be in extra instead'

    if needs_modify:
        import copy
        item = copy.deepcopy(item)
        item['properties']['config'] = config
        item['properties']['args'] = args

    return item


@xdev.profile
def load_pxl_eval(fpath, expt_dvc_dpath=None, arg_prefix='', mode=0, with_param_types=True):
    from kwcoco.coco_evaluator import CocoSingleResult
    # from watch.utils import result_analysis
    # from watch.utils import util_time
    measure_info = _load_json(fpath)

    meta = measure_info['meta']

    pred_info = meta['info']
    dvc_dpath = expt_dvc_dpath

    salient_measures = measure_info['nocls_measures']
    class_measures = measure_info['ovr_measures']

    # HACK: fixme
    coi_pattern = util_pattern.MultiPattern.coerce(
        ['Active Construction', 'Site Preparation'], hint='glob')

    class_metrics = []
    coi_metrics = []
    for catname, bin_measure in class_measures.items():
        class_row = {}
        class_row['AP'] = bin_measure['ap']
        class_row['AUC'] = bin_measure['auc']
        class_row['APUC'] = np.nanmean([bin_measure['ap'], bin_measure['auc']])
        class_row['catname'] = catname
        if coi_pattern.match(catname):
            coi_metrics.append(class_row)
        class_metrics.append(class_row)

    class_aps = [r['AP'] for r in class_metrics]
    class_aucs = [r['AUC'] for r in class_metrics]
    coi_aps = [r['AP'] for r in coi_metrics]
    coi_aucs = [r['AUC'] for r in coi_metrics]
    coi_catnames = [r['catname'] for r in coi_metrics]

    metrics = {}
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', 'Mean of empty slice')
        metrics['class_mAP'] = np.nanmean(class_aps) if len(class_aps) else np.nan
        metrics['class_mAUC'] = np.nanmean(class_aucs) if len(class_aucs) else np.nan
        metrics['class_mAPUC'] = np.nanmean([metrics['class_mAUC'], metrics['class_mAP']])

        metrics['coi_mAP'] = np.nanmean(coi_aps) if len(coi_aps) else np.nan
        metrics['coi_mAUC'] = np.nanmean(coi_aucs) if len(coi_aucs) else np.nan
        metrics['coi_mAPUC'] = np.nanmean([metrics['coi_mAUC'], metrics['coi_mAP']])

        metrics['salient_AP'] = salient_measures['ap']
        metrics['salient_AUC'] = salient_measures['auc']
        metrics['salient_APUC'] = np.nanmean([metrics['salient_AP'], metrics['salient_AUC']])

    for class_row in class_metrics:
        metrics[class_row['catname'] + '_AP'] = class_row['AP']
        metrics[class_row['catname'] + '_AUC'] = class_row['AUC']

    result = CocoSingleResult.from_json(measure_info)
    json_info = measure_info.copy()

    if with_param_types:
        param_types = parse_pred_pxl_params(pred_info, dvc_dpath, arg_prefix=arg_prefix, mode=mode)

        if mode == 0:
            predict_args = param_types[arg_prefix + 'pxl']
        else:
            predict_args = None

        if predict_args is None:
            raise Exception('no prediction metadata')

        HACK_HANDLE_CROPPED_AND_TRACK_PARAMS = 1
        if HACK_HANDLE_CROPPED_AND_TRACK_PARAMS and arg_prefix == 'act.':
            try:
                trk_param_types, extra_attrs = _handle_crop_and_trk_params(
                    param_types, expt_dvc_dpath)
                param_types.update(trk_param_types)
            except Exception:
                extra_attrs = {}
        else:
            extra_attrs = {}
        extra_attrs.update(_add_prefix(arg_prefix + 'pxl.metrics.', metrics))
        # quick and dirty way to get access to single-region results
        # This is not robust, done because it wasnt clear how to get
        # the equivalent test dataset for polygon eval variants.
        json_info['region_ids'] = ub.Path(predict_args['pxl.test_dataset']).name.split('.')[0]
    else:
        extra_attrs = {}
        param_types = None

    info = {
        'fpath': fpath,
        'metrics': metrics,
        'param_types': param_types,
        'other': {
            'result': result,
            'extra_attrs': extra_attrs,
            'coi_catnames': ','.join(sorted(coi_catnames)),
            # 'sc_cm': sc_cm,
            # 'sc_df': sc_df,
        },
        'json_info': json_info,
    }

    return info


class Found(Exception):
    pass


@xdev.profile
def resolve_cross_machine_path(path, dvc_dpath=None):
    """
    HACK

    Attempt to determine what the local path to a file/directry would be
    if it exists on this machine. This assumes the path is something
    that was checked into DVC.

    Args:
        dvc_dpath : the preferred dvc dpath to associate the file with
            in case the older one points to multiple.
    """
    # pkg_dvc = SimpleDVC.find_root(package_fpath).resolve()
    # SimpleDVC.find_root(package_fpath).resolve()
    path = ub.Path(path)
    needs_resolve = not path.exists()
    if not needs_resolve:
        if dvc_dpath is not None:
            needs_resolve = not path.is_relative_to(dvc_dpath)

    if needs_resolve:
        expected_dnames = [
            'smart_watch_dvc',
            'smart_watch_dvc-hdd',
            'smart_watch_dvc-ssd',
        ]

        found_idx = None
        for dname in expected_dnames:
            try:
                idx = path.parts.index(dname)
            except ValueError:
                pass
            else:
                found_idx = idx
                break

        if found_idx is not None:
            # import watch
            # dvc_dpath = watch.find_smart_dvc_dpath()
            pname = ub.Path(*path.parts[idx + 1:])
            pname_dvc = pname.augment(tail='.dvc')
            cwd = ub.Path('.').absolute()
            candidates = []
            if dvc_dpath is not None:
                candidates.extend([
                    dvc_dpath / pname,
                    dvc_dpath / pname_dvc,
                ])
            candidates.extend([
                cwd / pname,
                cwd / pname_dvc,
            ])
            found = None
            try:
                for cand_path in candidates:
                    if cand_path.exists():
                        found = cand_path
                        if found.name.endswith('.dvc'):
                            found = found.augment(ext='')
                        raise Found
            except Found:
                pass
            if found:
                return found
    return path


@ub.memoize
def global_ureg():
    import pint
    ureg = pint.UnitRegistry()
    return ureg


global_ureg()


def _add_prefix(prefix, dict_):
    return {prefix + k: v for k, v in dict_.items()}


@xdev.profile
def relevant_pred_pxl_config(pred_pxl_config, dvc_dpath=None, arg_prefix=''):
    # TODO: better way of inferring what params are relevant
    # This should be metadata a scriptconfig object can hold.
    pred_config = {}
    pred_config['tta_fliprot'] = pred_pxl_config.get('tta_fliprot', 0)
    pred_config['tta_time'] = pred_pxl_config.get('tta_time', 0)
    pred_config['chip_overlap'] = pred_pxl_config['chip_overlap']
    pred_config['input_space_scale'] = pred_pxl_config.get('input_space_scale', None)
    pred_config['window_space_scale'] = pred_pxl_config.get('window_space_scale', None)
    pred_config['output_space_scale'] = pred_pxl_config.get('output_space_scale', None)
    pred_config['time_span'] = pred_pxl_config.get('time_span', None)
    pred_config['time_sampling'] = pred_pxl_config.get('time_sampling', None)
    pred_config['time_steps'] = pred_pxl_config.get('time_steps', None)
    pred_config['chip_dims'] = pred_pxl_config.get('chip_dims', None)
    pred_config['set_cover_algo'] = pred_pxl_config.get('set_cover_algo', None)
    pred_config['resample_invalid_frames'] = pred_pxl_config.get('resample_invalid_frames', None)
    pred_config['use_cloudmask'] = pred_pxl_config.get('use_cloudmask', None)
    package_fpath = pred_pxl_config['package_fpath']
    test_dataset = pred_pxl_config['test_dataset']
    if dvc_dpath is not None:
        package_fpath = resolve_cross_machine_path(package_fpath, dvc_dpath)
        test_dataset = resolve_cross_machine_path(test_dataset, dvc_dpath)

    # pred_config['model_fpath'] = package_fpath
    # pred_config['in_dataset'] = test_dataset
    pred_config['package_fpath'] = package_fpath
    pred_config['test_dataset'] = test_dataset

    # FIXME: use a common heuristic to transform a path into a model.
    test_dataset = ub.Path(test_dataset)
    pred_config['properties.model_name'] = model_name = ub.Path(package_fpath).name
    pred_config['properties.dataset_name'] = str(ub.Path(*test_dataset.parts[-2:]))

    # Hack to get the epoch/step/expt_name
    try:
        epoch = int(model_name.split('epoch=')[1].split('-')[0])
    except Exception:
        epoch = -1
    try:
        step = int(model_name.split('step=')[1].split('-')[0])
    except Exception:
        step = -1
    try:
        expt_name = model_name.split('_epoch=')[0]
    except Exception:
        expt_name = '?'
        # expt_name = predict_args[expt_name]

    pred_config['properties.step'] = step
    pred_config['properties.epoch'] = epoch
    pred_config['properties.expt_name'] = expt_name
    pred_config = _add_prefix(arg_prefix + 'pxl.', pred_config)
    return pred_config


@xdev.profile
def relevant_fit_config(fit_config, arg_prefix='', add_prefix=True):
    ignore_params = {
        'default_root_dir', 'enable_progress_bar'
        'prepare_data_per_node', 'enable_model_summary', 'checkpoint_callback',
        'detect_anomaly', 'gpus', 'terminate_on_nan', 'train_dataset',
        'workdir', 'config', 'num_workers', 'amp_backend',
        'enable_progress_bar', 'flush_logs_every_n_steps',
        'enable_checkpointing', 'prepare_data_per_node', 'amp_level',
        'vali_dataset', 'test_dataset', 'package_fpath', 'num_draw',
        'num_nodes', 'num_processes', 'num_sanity_val_steps',
        'overfit_batches', 'process_position',
        'reload_dataloaders_every_epoch', 'reload_dataloaders_every_n_epochs',
        'replace_sampler_ddp', 'sync_batchnorm', 'torch_sharing_strategy',
        'torch_start_method', 'val_check_interval', 'weights_summary',
        'auto_lr_find', 'auto_select_gpus', 'auto_scale_batch_size', 'benchmark',
        'check_val_every_n_epoch', 'draw_interval', 'eval_after_fit', 'fast_dev_run',
        'limit_predict_batches', 'limit_test_batches', 'limit_train_batches',
        'limit_val_batches', 'log_every_n_steps', 'logger',
        'move_metrics_to_cpu', 'multiple_trainloader_mode',
    }
    from scriptconfig import smartcast
    # hack, rectify different values of known parameters that mean the
    # same thing
    fit_config2 = ub.dict_diff(fit_config, ignore_params)
    for k, v in fit_config2.items():
        if k not in {'channels', 'init'}:
            v2 = smartcast.smartcast(v)
            if isinstance(v2, list):
                # Dont coerce into a list
                v2 = v
            fit_config2[k] = v2
        else:
            fit_config2[k] = v

    if 'init' in fit_config2:
        # hack to make init only use the filename
        fit_config2['init'] = fit_config2['init'].split('/')[-1]
    if add_prefix:
        fit_config2 = _add_prefix(arg_prefix + 'fit.', fit_config2)
    return fit_config2


@xdev.profile
def relevant_track_config(track_args, arg_prefix=''):
    track_config = json.loads(track_args['track_kwargs'])
    track_config = _add_prefix(arg_prefix + 'poly.', track_config)
    return track_config


@xdev.profile
def parse_resource_item(item, arg_prefix='', add_prefix=True):
    resources = {}

    ureg = global_ureg()
    pred_prop = item['properties']

    start_time = util_time.coerce_datetime(pred_prop.get('start_timestamp', None))
    end_time = util_time.coerce_datetime(pred_prop.get('end_timestamp', pred_prop.get('stop_timestamp', None)))
    iters_per_second = pred_prop.get('iters_per_second', None)
    if start_time is None or end_time is None:
        total_hours = None
    else:
        total_hours = (end_time - start_time).total_seconds() / (60 * 60)
    resources['total_hours'] = total_hours
    if iters_per_second is not None:
        resources['iters_per_second'] = iters_per_second

    if 'duration' in pred_prop:
        resources['duration'] = pred_prop['duration']

    try:
        vram = pred_prop['device_info']['allocated_vram']
        vram_gb = ureg.parse_expression(f'{vram} bytes').to('gigabytes').m
        resources['vram_gb'] = vram_gb
    except KeyError:
        ...

    hardware_parts = []

    if 'machine' in pred_prop:
        cpu_name = pred_prop['machine']['cpu_brand']
        cpu_name = re.sub('.*Gen Intel.R. Core.TM. ', '', cpu_name)
        resources['cpu_name'] = cpu_name
        hardware_parts.append(cpu_name)

    try:
        gpu_name = pred_prop['device_info']['device_name']
        resources['gpu_name'] = gpu_name
        hardware_parts.append(gpu_name)
    except KeyError:
        ...

    if 'emissions' in pred_prop:
        co2_kg = pred_prop['emissions']['co2_kg']
        kwh = pred_prop['emissions']['total_kWH']
        resources['co2_kg'] = co2_kg
        resources['kwh'] = kwh

    if 'disk_info' in pred_prop:
        disk_type = pred_prop['disk_info']['filesystem']
        resources['disk_type'] = disk_type

    resources['hardware'] = ' '.join(hardware_parts)
    if add_prefix:
        resources = _add_prefix(arg_prefix + 'resource.', resources)
    return resources


@xdev.profile
def find_pred_pxl_item(pred_info):
    pred_items = list(find_info_items(
        pred_info,
        {'process', 'process_context'},
        'watch.tasks.fusion.predict'
    ))
    assert len(pred_items) == 1
    pred_item = pred_items[0]
    return pred_item


@xdev.profile
def find_info_items(info, query_type, query_name=None):
    from watch.utils import util_pattern
    if query_name is None:
        query_name = '*'
    query_name_pattern = util_pattern.MultiPattern.coerce(query_name)
    query_type_pattern = util_pattern.MultiPattern.coerce(query_type)
    for item in info:
        if query_type_pattern.match(item['type']):
            name = item['properties']['name']
            if query_name_pattern.match(name):
                yield item


@xdev.profile
def parse_pred_pxl_params(pred_info, expt_dvc_dpath=None, arg_prefix='', mode=0):
    """
    Args:
        pred_info (List[Dict]):
            the info written to a heatmap prediction kwcoco
    """
    pred_item = find_pred_pxl_item(pred_info)
    pred_item = _handle_process_item(pred_item)

    # NOTE: the place where measure are stored has changed to be inside
    # the pred item.
    assert not list(find_info_items(pred_info, 'measure', None))

    meta = {'start_time': None}
    resources = {}
    # New code should have measures inside the pred item
    fit_config = pred_item['properties']['extra']['fit_config']
    meta['start_time'] = pred_item['properties']['start_timestamp']
    meta['end_time'] = pred_item['properties']['stop_timestamp']
    meta['duration'] = pred_item['properties']['duration']
    meta['uuid'] = pred_item['properties']['uuid']
    # meta['start_timestamp'] = pred_item['properties']['start_timestamp']
    # meta['stop_timestamp'] = pred_item['properties']['stop_timestamp']

    resources = parse_resource_item(pred_item, arg_prefix=(arg_prefix + 'pxl.'))

    pred_pxl_config = pred_item['properties']['config']
    pred_pxl_config = relevant_pred_pxl_config(
        pred_pxl_config, expt_dvc_dpath, arg_prefix=arg_prefix)
    fit_config = relevant_fit_config(fit_config, arg_prefix=arg_prefix)

    if mode:
        param_types = {
            arg_prefix + 'fit': fit_config,
            arg_prefix + 'pxl': pred_pxl_config,
            arg_prefix + 'pxl.resource': resources,
            arg_prefix + 'pxl.meta': _add_prefix(arg_prefix + 'pxl.meta.', meta),
        }
    else:
        param_types = {
            arg_prefix + 'fit': fit_config,
            arg_prefix + 'pxl': pred_pxl_config,
            arg_prefix + 'pxl.resource': resources,
            arg_prefix + 'pxl.meta': _add_prefix(arg_prefix + 'pxl.meta.', meta),
        }
    return param_types


@ub.memoize
def _load_json(fpath):
    # memo hack for development
    with open(fpath, 'r') as file:
        data = json.load(file)
    return data


@xdev.profile
def find_track_item(tracker_info):
    tracker_alias = {
        'watch.cli.kwcoco_to_geojson',
        'watch.cli.run_tracker',
    }
    track_items = list(find_info_items(
        tracker_info,
        {'process', 'process_context'},
        tracker_alias
    ))
    if len(track_items) != 1:
        raise AssertionError(ub.paragraph(
            f'''
            We should be able to find exactly 1 tracker process item,
            but instead we found {len(track_items)}
            '''))
        ...
    track_item = track_items[0]
    return track_item


@xdev.profile
def find_metrics_framework_item(info):
    task_aliases = {
        'watch.cli.run_metrics_framework',
    }
    items = list(find_info_items(
        info,
        {'process', 'process_context'},
        task_aliases
    ))
    if len(items) != 1:
        raise AssertionError(ub.paragraph(
            f'''
            We should be able to find exactly 1 tracker process item,
            but instead we found {len(items)}
            '''))
        ...
    item = items[0]
    return item


@xdev.profile
def find_pxl_eval_item(info):
    task_aliases = {
        'watch.tasks.fusion.evaluate',
    }
    items = list(find_info_items(
        info,
        {'process', 'process_context'},
        task_aliases
    ))
    if len(items) != 1:
        raise AssertionError(ub.paragraph(
            f'''
            We should be able to find exactly 1 tracker process item,
            but instead we found {len(items)}
            '''))
        ...
    item = items[0]
    return item


@xdev.profile
def shrink_channels(x):
    import kwcoco
    aliases = {
        'blue': 'B',
        'red': 'R',
        'green': 'G',
        'nir': 'N',
        'swir16': 'S',
        'swir22': 'H',
    }
    for idx, part in enumerate('forest|brush|bare_ground|built_up|cropland|wetland|water|snow_or_ice_field'.split('|')):
        aliases[part] =  'land.{}'.format(idx)
    stream_parts = []
    sensorchan = kwcoco.SensorChanSpec.coerce(x)
    # spec = kwcoco.ChannelSpec.coerce(x)
    for stream in sensorchan.streams():
        fused_parts = []
        for c in stream.chans.as_list():
            c = aliases.get(c, c)
            c = c.replace('matseg_', 'matseg.')
            fused_parts.append(c)
        fused = '|'.join(fused_parts)
        fused = fused.replace('B|G|R|N', 'BGRN')
        fused = fused.replace('B|G|R|N|S|H', 'BGRNSH')
        fused = fused.replace('R|G|B', 'RGB')
        fused = fused.replace('B|G|R', 'BGR')
        stream_parts.append(stream.sensor.spec + ':' + fused)
    new = ','.join(stream_parts)
    x = kwcoco.SensorChanSpec.coerce(new).concise().spec
    return x


@xdev.profile
def is_teamfeat(sensorchan):
    """
    Check if the sensorchan spec contains a hard coded value we know is a team
    feature
    """
    import math
    unique_chans = sum([s.chans for s in sensorchan.streams()]).fuse().to_set()
    if isinstance(unique_chans, float) and math.isnan(unique_chans):
        return False
    return any([a in unique_chans for a in ['depth', 'invariant', 'invariants', 'matseg', 'land']])



r"""
This is the CLI for expt_state

Synchronize DVC states across the machine.

This is a new Phase2 Variant of this script.

Example:

    export DVC_DATA_DPATH=$(geowatch_dvc --tags="phase2_data")
    export DVC_EXPT_DPATH=$(geowatch_dvc --tags="phase2_expt")
    cd $DVC_EXPT_DPATH

    python -m watch mlops "status" --dataset_codes "Aligned-Drop4-2022-08-08-TA1-S2-WV-PD-ACC"
    python -m watch mlops "status" --dataset_codes "Aligned-Drop4-2022-08-08-TA1-S2-WV-PD-ACC"
    python -m watch mlops "pull packages evals" --dataset_codes "Aligned-Drop4-2022-08-08-TA1-S2-WV-PD-ACC"
    python -m watch mlops "push packages evals"

    python -m watch mlops "status" --dataset_codes Drop4-SC

    python -m watch mlops "list"

    # On training machine
    python -m watch mlops "push packages"
    python -m watch mlops "push packages" --dataset_codes "Aligned-Drop4-2022-08-08-TA1-S2-WV-PD-ACC"

    # On testing machine
    python -m watch mlops "pull packages"
    python -m watch mlops "status"

    # Run evals on testing machine
    python -m watch mlops "evaluate" --dataset_codes "Aligned-Drop4-2022-08-08-TA1-S2-WV-PD-ACC"

    # On testing machine
    python -m watch mlops "push evals"

    # On analysis machine
    python -m watch mlops "pull evals"


TODO:
    ### Make the Experiment Evaluation Reporter more robust and generalize to
    ### more problems.

    It should quickly show the best models for various metric and it should be
    easy for the user to inspect them further.  For example say the best model
    of interest was:

    MODEL_OF_INTEREST="Drop4_BAS_Retrain_V002_epoch=45-step=23552"
    MODEL_OF_INTEREST="Drop4_BAS_Continue_15GSD_BGR_V004_epoch=78-step=323584"

    # TODO:
    # There is a problem with multiple .pt suffixes, just dont use any

    # You should be able to pull things wrt to that model

    python -m watch mlops "pull packages" --model_pattern="${MODEL_OF_INTEREST}*"
    python -m watch mlops "pull evals" --model_pattern="${MODEL_OF_INTEREST}*"
    python -m watch mlops "status" --model_pattern="${MODEL_OF_INTEREST}*"

Ignore:
    python -m watch mlops "evaluate" \
        --enable_pred=1 \
        --enable_eval=1 \
        --enable_actclf=1 \
        --enable_actclf_eval=1 \
        --dataset_codes "Aligned-Drop4-2022-08-08-TA1-S2-WV-PD-ACC" \
        --devices="0,1,2,3" --run=1

    python -m watch mlops "evaluate" \
        --bas_thresh=0.0,0.01,0.1 \
        --set_cover_algo=approx,exact \
        --enable_pred=1 \
        --enable_eval=1 \
        --enable_track=1 \
        --enable_iarpa_eval=1 \
        --dataset_codes "Aligned-Drop4-2022-08-08-TA1-S2-L8-ACC" \
        --devices="0,1" --run=0 \
        --skip_existing=True

        # --enable_track=1 \
        # --enable_iarpa_eval=1 \
"""
import pandas as pd
import ubelt as ub
import platform
import scriptconfig as scfg
from watch.utils import simple_dvc
from watch import heuristics
from watch.mlops.old.expt_state import ExperimentState, summarize_tables


class ExptManagerConfig(scfg.DataConfig):
    """
    Certain parts of these names have special nomenclature to make them easier
    to work with in Python and Bash.

    The "watch" module comes with a few nice command line programs. Given a
    machine with the "watch" environment, the watch DVC repo is accessed as
    follows:

        DVC_EXPT_DPATH=$(geowatch_dvc --tags="phase2_expt")
        DVC_DATA_DPATH=$(geowatch_dvc --tags="phase2_data")

    The workdir is where a user on a machine puts all of their experiments.

        WORKDIR=$DVC_EXPT_DPATH/training/$HOSTNAME/$USER

    Before we start an experment, we must choose a dataset. Lets use an
    example:

        DATASET_CODE=Aligned-Drop3-L1

    Along with the DVC directory, this should uniquely specify a kwcoco dataset
    bundle (although it might not specify the specific view of that dataset,
    -- views can have different GSD or be bundle subsets). The directory
    of this bundle should be:

        KWCOCO_BUNDLE_DPATH=$DVC_DATA_DPATH/$DATASET_CODE

    and it should be the case that, there is a kwcoco manifest that describes
    the entire bundle called:

        MAIN_FPATH=$KWCOCO_BUNDLE_DPATH/data.kwcoco.json

    But we may have precomputed splits also e.g

        TRAIN_FPATH=$KWCOCO_BUNDLE_DPATH/data_nowv_train.kwcoco.json
        VALI_FPATH=$KWCOCO_BUNDLE_DPATH/data_nowv_vali.kwcoco.json

    Each experiment should also be given a name:

        EXPERIMENT_NAME=Drop3_L1_BASELINE_BAS_V001

    For each experiment you choose an experiment name on a dataset.

        DEFAULT_ROOT_DIR=$WORKDIR/$DATASET_CODE/runs/$EXPERIMENT_NAME

    The packaging part of this script works with

        $DVC_EXPT_DPATH/models/fusion/$EXPT_GROUP_CODE/packages/$EXPT_MODEL_GLOBNAME/*.pt
    """
    command = scfg.Value(None, nargs='*', help='if specified, will overload other options', position=1)

    dvc_remote = scfg.Value('aws', help='dvc remote to sync to/from')

    expt_dvc_dpath = scfg.Value('auto', help='path to the experiment dvc dpath')
    data_dvc_dpath = scfg.Value('auto', help='path to the data dvc dpath')

    model_pattern = scfg.Value('*', help='if specified restrict to models matching this name pattern')

    dataset_codes = scfg.Value(None, nargs='+', help=ub.paragraph(
        '''
        if unset, will use the defaults, otherwise this should be a list of
        the DVC dataset bundle names that we want to consider.  Note: we do
        make assumptions that the about where these names go in paths.
        We may make this more general in the future.

        Namely:

            # Training runs go here go here.
            <expt_dvc_dpath>/training/*/*/<dataset_code>/runs/<expt_name>/lightning_logs

            # Packages go here.
            <expt_dvc_dpath>/models/fusion/<dataset_code>/packages

            # Evaluations go here.
            <expt_dvc_dpath>/models/fusion/<dataset_code>/eval

        NOTE: THIS SPECIFIC FORMAT IS IN HIGH FLUX. DOCS MAY BE OUTDATED
        '''))


def main(cmdline=True, **kwargs):
    """
    from watch.mlops.old.expt_manager import *  # NOQA
    """
    config = ExptManagerConfig(cmdline=cmdline, data=kwargs)
    print('ExptManagerConfig config = {}'.format(ub.urepr(dict(config), nl=1)))
    command = config['command']

    available_actions = [
        'status', 'evaluate', 'push', 'pull', 'list', 'report',
    ]
    available_targets = [
        'packages', 'evals'
    ]

    actions = []
    targets = []

    if command is not None and command:
        for c in command:
            for a in available_actions:
                if a in c:
                    actions.append(a)
            for t in available_targets:
                if t in c:
                    targets.append(t)

    # print(f'actions={actions}')
    # print(f'targets={targets}')
    # print('config = {}'.format(ub.urepr(dict(config), nl=1)))

    dvc_remote = config['dvc_remote']

    if config['dataset_codes'] is None:
        dataset_codes = heuristics.DATASET_CODES
    else:
        dataset_codes = config['dataset_codes']

    if config['expt_dvc_dpath'] == 'auto':
        config['expt_dvc_dpath'] = heuristics.auto_expt_dvc()
    if config['data_dvc_dpath'] == 'auto':
        config['data_dvc_dpath'] = heuristics.auto_data_dvc()

    expt_dvc_dpath = config['expt_dvc_dpath']
    data_dvc_dpath = config['data_dvc_dpath']

    manager = DVCExptManager(
        expt_dvc_dpath, dvc_remote=dvc_remote, dataset_codes=dataset_codes,
        data_dvc_dpath=data_dvc_dpath, model_pattern=config['model_pattern'])

    if 'pull' in actions:
        manager.pull(targets)

    if 'push' in actions:
        manager.push(targets)

    if 'status' in actions:
        manager.summarize()

    if 'list' in actions:
        manager.list()

    if 'evaluate' in actions:
        self = manager
        for state in self.states:
            state.schedule_evaluation()

    if 'report' in actions:
        self = manager
        from watch.mlops import expt_report
        dvc_manager = manager
        reporter = expt_report.EvaluationReporter(dvc_manager)
        reporter.load()
        reporter.status()
        reporter.plot()


class DVCExptManager(ub.NiceRepr):
    """
    Implements an API around our DVC structure, which can be described as
    follows.

    TODO:
        - [ ] If we can somehow generate the output paths based on the
        pipeline, then we will be in a very good position.

    <data_dvc_dpath>
        * [<dataset_code>, ...]

    <expt_dvc_dpath>
        * training
            * <hostname>/<user>/<dataset_code>/runs/<expt_name>/lightning_logs/...

        * models
            * <task>
            * fusion/<dataset_code>/packages/<expt_name>/<model_name.pt>
            * fusion/<dataset_code>/pred/<expt_name>/<model_name.pt>/***
            * fusion/<dataset_code>/eval/<expt_name>/<model_name.pt>/***

    A breakdown of the packages dir is:
        packages/<expt_name>/<model_name.pt>

    Example:
        >>> # xdoctest: +REQUIRES(env:DVC_EXPT_DPATH)
        >>> from watch.mlops.old.expt_manager import *  # NOQA
        >>> import watch
        >>> manager = DVCExptManager.coerce(watch.find_dvc_dpath(tags='phase2_expt'))
        >>> manager.summarize()

    Ignore:
        broke = df[df['is_broken']]
    """

    def __nice__(manager):
        return str(manager.dvc)

    def __init__(manager, expt_dvc_dpath, dvc_remote='aws', dataset_codes=None,
                 data_dvc_dpath=None, model_pattern='*'):
        manager.model_pattern = model_pattern
        manager.expt_dvc_dpath = expt_dvc_dpath
        manager.data_dvc_dpath = data_dvc_dpath
        manager.dvc_remote = dvc_remote
        manager.dataset_codes = dataset_codes
        manager.dvc = simple_dvc.SimpleDVC.coerce(expt_dvc_dpath, remote=dvc_remote)
        manager._build_states()

    def summarize(manager):
        # for state in manager.states:
        #     state.summarize()
        tables = manager.cross_referenced_tables()
        summarize_tables(tables)

    def list(manager):
        tables = manager.cross_referenced_tables()

        if 'staging' in tables:
            todrop = ['expt_dvc_dpath', 'raw', 'ckpt_path', 'spkg_fpath', 'pkg_fpath', 'lightning_version', 'ckpt_exists']
            df = tables['staging']
            print(df.drop(ub.oset(todrop) & df.columns, axis=1).to_string())

        if 'volitile' in tables:
            volitile_drop = ['expt_dvc_dpath', 'raw', '']
            todrop = volitile_drop
            df = tables['volitile']
            print(df.drop(ub.oset(todrop) & df.columns, axis=1).to_string())

        if 'versioned' in tables:
            volitile_drop = ['raw', 'dvc', 'expt_dvc_dpath', 'expt',
                             'is_broken', 'is_link', 'has_raw', 'has_dvc',
                             'unprotected', 'needs_pull', 'needs_push',
                             'has_orig']
            todrop = volitile_drop
            df = tables['versioned']
            type_to_versioned = dict(list(df.groupby('type')))
            for type, subdf in type_to_versioned.items():
                print(f'type={type}')
                print(subdf.drop(ub.oset(todrop) & df.columns, axis=1).to_string())

    @classmethod
    def coerce(cls, expt_dvc_dpath=None):
        import watch
        if expt_dvc_dpath is None:
            expt_dvc_dpath = watch.find_smart_dvc_dpath()
        dvc_remote = 'aws'
        dataset_codes = heuristics.DATASET_CODES
        manager = cls(expt_dvc_dpath=expt_dvc_dpath, dvc_remote=dvc_remote,
                      dataset_codes=dataset_codes)
        return manager

    def _build_states(manager):
        states = []
        for dataset_code in manager.dataset_codes:
            state = ExperimentState(
                manager.expt_dvc_dpath, dataset_code, dvc_remote=manager.dvc_remote,
                data_dvc_dpath=manager.data_dvc_dpath,
                model_pattern=manager.model_pattern)
            states.append(state)
        manager.states = states

    def versioned_table(manager, **kw):
        rows = list(ub.flatten(state.versioned_rows(**kw) for state in manager.states))
        df = pd.DataFrame(rows)
        missing = ub.oset(ExperimentState.VERSIONED_COLUMNS) - df.columns
        if len(missing):
            df.loc[:, missing] = None
        return df

    def volitile_table(manager, **kw):
        rows = list(ub.flatten(state.volitile_table(**kw) for state in manager.states))
        df = pd.DataFrame(rows)
        missing = ub.oset(ExperimentState.VOLITILE_COLUMNS) - df.columns
        if len(missing):
            df.loc[:, missing] = None
        return df

    def evaluation_table(manager):
        rows = list(ub.flatten(state.evaluation_rows() for state in manager.states))
        df = pd.DataFrame(rows)
        return df

    def cross_referenced_tables(manager):
        import pandas as pd
        table_accum = ub.ddict(list)
        for state in manager.states:
            tables = state.cross_referenced_tables()
            for k, v in tables.items():
                if not v.empty:
                    # col 'n_pred_act_poly_sites_fpath' can be duplicated
                    table_accum[k].append(
                        v.loc[:, ~v.columns.duplicated()]
                    )
        combo_tables = ub.udict(table_accum).map_values(lambda vs: pd.concat(vs))
        return combo_tables

    def push_evals(manager):
        dvc = manager.dvc
        eval_df = manager.evaluation_table()
        summarize_tables({'versioned': eval_df})

        is_weird = (eval_df.is_link & (~eval_df.has_dvc))
        weird_df = eval_df[is_weird]
        if len(weird_df):
            print(f'weird_df=\n{weird_df}')

        # Determine what evaluations need to be added to DVC
        to_add = eval_df[~eval_df['has_dvc']]
        if len(to_add):
            to_add_paths = to_add['raw']
            # paths = to_add_paths
            dvc.add(to_add_paths)
            ...

        to_push = eval_df[eval_df.needs_push == True]  # NOQA
        assert not to_push['has_dvc'].any()

        # TODO: if we want to allow modifications we need to find
        # unprotected files (or changed files on non-symlink dvc repos)
        # to_push = eval_df[(eval_df.needs_push == True) | (eval_df.unprotected == True)]  # NOQA

        to_push_fpaths = to_push['raw'].tolist()
        print(f'to_push=\n{to_push}')
        if len(to_push_fpaths):
            # dvc.add(to_push_fpaths)
            dvc.git_commitpush(f'Sync evals from {platform.node()}')
            dvc.push(to_push_fpaths)

    def pull_evals(manager):
        manager.dvc.git_pull()

        eval_df = manager.evaluation_table()
        summarize_tables({'versioned': eval_df})

        # manager.summarize()
        print(f'manager.expt_dvc_dpath={manager.expt_dvc_dpath}')
        print(len(eval_df))
        if len(eval_df) > 0:
            eval_df = eval_df[~eval_df['is_broken']]
            pull_rows = eval_df[eval_df.needs_pull]
            pull_fpaths = pull_rows['dvc'].tolist()
        else:
            pull_fpaths = []
        print(f'{len(pull_fpaths)=}')
        for p in pull_fpaths:
            assert p.exists()
        manager.dvc.pull(pull_fpaths)

    def pull_packages(manager):
        # Assume just one git repo and manually pull
        manager.dvc.git_pull()

        pkg_df = manager.versioned_table(types=['pkg_fpath'])
        pull_df = pkg_df[pkg_df['needs_pull'].astype(bool)]
        pull_fpaths = pull_df['dvc'].tolist()
        manager.dvc.pull(pull_fpaths)

    def push_packages(manager):
        """
        TODO: break this up into smaller components.
        """
        # from watch.tasks.fusion import repackage
        # mode = 'commit'
        for state in manager.states:
            state.push_packages()

    def push(manager, targets):
        if 'packages' in targets:
            manager.push_packages()

        if 'evals' in targets:
            manager.push_evals()

    def pull(manager, targets):
        if 'packages' in targets:
            manager.pull_packages()
        if 'evals' in targets:
            manager.pull_evals()

    def reverse_hash_lookup(manager, key):
        # This probably doesn't belong here
        from watch.utils.reverse_hashid import ReverseHashTable
        ReverseHashTable.query(key, verbose=1)


# TODO: can we hook into DVC more efficiently to query this?
# def _check_ignore_tables(paths, dvc):
#     import os
#     dvc_root = dvc.dvc_root
#     @ub.memoize
#     def make_gitignore_pattern(root):
#         from watch.utils import util_pattern
#         # ignore_fpath = (root / '.gitignore')
#         ignore_fpath = (root / '.dvcignore')
#         if ignore_fpath.exists():
#             ignore_lines = [p.strip() for p in ignore_fpath.read_text().split('\n')
#                             if not p.startswith('#')]
#             ignore_pats = [str(p) for p in ignore_lines if p]
#             pat = util_pattern.MultiPattern.coerce(ignore_pats, hint='glob')
#             return pat
#     for path in ub.ProgIter(paths):
#         rel_path = path.relative_to(dvc_root).parent
#         for i in reversed(range(len(rel_path.parts))):
#             rel_root = dvc_root / ub.Path(*rel_path.parts[0:i])
#             rel_pat = make_gitignore_pattern(rel_root)
#             if rel_pat is not None:
#                 print(f'rel_root={rel_root}')
#                 suffix_path = os.fspath(path.relative_to(rel_root))
#                 if rel_pat.match(suffix_path):
#                     raise Exception


__config__ = ExptManagerConfig


if __name__ == '__main__':
    """
    CommandLine:
        python ~/code/watch/watch/mlops/expt_manager.py "pull all"
        python -m watch.mlops.expt_manager "status"
        python -m watch.mlops.expt_manager "push all"
        python -m watch.mlops.expt_manager "pull evals"

        python -m watch.mlops.expt_manager "evaluate"

        python -m watch.mlops.expt_manager "pull all"
        python -m watch.mlops.expt_manager "pull packages"
    """
    main(cmdline=True)



import kwcoco
# import math
import numpy as np
import pandas as pd
import ubelt as ub
from watch import heuristics

fit_param_keys = heuristics.fit_param_keys
pred_param_keys = heuristics.pred_param_keys
trk_param_keys = heuristics.trk_param_keys
act_param_keys = heuristics.act_param_keys
DSET_CODE_TO_GSD = heuristics.DSET_CODE_TO_GSD


class UnableToPlot(Exception):
    pass


class Plotter:
    """
    A WIP refactor to hold the information we need to do plotting.
    """
    def __init__(plotter):
        pass

    @classmethod
    def from_reporter(cls, reporter, common_plotkw, dpath):
        plotter = Plotter()
        orig_merged_df = reporter.orig_merged_df
        iarpa_metric_lut = reporter.iarpa_metric_lut
        pixel_metric_lut = reporter.pixel_metric_lut
        # predcfg_to_label = reporter.predcfg_to_label
        # actcfg_to_label = reporter.actcfg_to_label
        human_mapping = reporter.human_mapping
        merged_df = orig_merged_df.copy()
        plotter.dpath = dpath
        plotter.metric_luts = {
            'pxl': pixel_metric_lut,
            'trk': iarpa_metric_lut,
            'act': iarpa_metric_lut,
        }
        common_plotkw = common_plotkw
        plotter.common_plotkw = common_plotkw
        plotter.human_mapping = human_mapping
        plotter.merged_df = merged_df

        # It's important to separate results by what dataset they were tested
        # on / what type of result they were evaluating
        plotter.group_keys = ['test_trk_dset', 'type']
        plotter.expt_groups = dict(list(merged_df.groupby(plotter.group_keys)))
        return plotter

    def plot_groups(plotter, plot_name, **kwargs):
        """
        plot_name = 'metric_over_training_time'
        metrics = ['BAS_F1']
        """
        plot_method = getattr(plotter, plot_name)
        for code_type, group in plotter.expt_groups.items():
            print(f'code_type={code_type}')
            try:
                plot_method(code_type, group, **kwargs)
            except UnableToPlot as ex:
                print(f'ex={ex}')
                continue

    def metric_over_training_time(plotter, code_type, group, metrics, huevar='expt'):
        test_dset, type = code_type
        metrics = metrics if ub.iterable(metrics) else [metrics]
        metrics_key = '_'.join(metrics)
        plot_name = 'metric_over_time_' + metrics_key
        prefix = f'{test_dset}_{type}_'
        plotkw = ub.udict({
            'x': 'step',
            'y': 'value',
            'star': 'in_production',
        })

        missing = set((plotkw & {'x'}).values()) - set(group.columns)
        if missing:
            raise UnableToPlot(f'Cannot plot {plot_name} for {code_type} missing={missing}')

        missing = set((plotkw & {'x'}).values()) - set(group.columns)
        if missing:
            raise UnableToPlot(f'Cannot plot {plot_name} for {code_type} missing={missing}')

        if plotkw['x'] not in group.columns:
            raise UnableToPlot

        if all(m not in group.columns for m in metrics):
            raise UnableToPlot

        melted = group.melt(
            ['step', 'in_production', 'expt', 'pred_cfg'],
            metrics, var_name='metric')

        if len(metrics) > 1:
            plotkw['style'] = 'metric'

        plotkw['hue'] = huevar

        def make_fig(fnum, legend=True):
            import kwplot
            fig = kwplot.figure(fnum=fnum, doclf=True)
            ax = fig.gca()
            humanized_scatterplot(plotter.human_mapping, plot_type='line',
                                  data=melted, ax=ax, legend=0, **plotkw)
            humanized_scatterplot(plotter.human_mapping, plot_type='scatter',
                                  data=melted, ax=ax, legend=legend, s=80,
                                  **plotkw)
            if len(metrics) == 1:
                ax.set_ylabel(metrics[0])
            nice_type = plotter.human_mapping.get(type, type)
            ax.set_title(f'Metric over time - {nice_type} - {test_dset}')

        fnum = plot_name + prefix
        run_make_fig(make_fig, fnum, plotter.dpath, plotter.human_mapping,
                     plot_name, prefix)

    def plot_relationship(plotter, code_type, group, huevar='sensorchan'):
        import kwplot
        plot_name = 'plot_relationship'

        x = 'act.poly.metrics.micro_f1'
        y = 'trk.poly.metrics.bas_f1'

        test_dset, type = code_type
        plotkw = ub.udict({
            # 'x': plotter.metric_luts['pxl'][type],
            # 'y': plotter.metric_luts['trk'][type],
            'x': x,
            'y': y,
            'hue': huevar,
            **plotter.common_plotkw,
        })

        missing = set((plotkw & {'x', 'y'}).values()) - set(group.columns)
        if missing:
            raise UnableToPlot(f'Cannot plot {plot_name} for {code_type} missing={missing}')

        if plotkw['x'] not in group.columns or plotkw['y'] not in group.columns:
            raise UnableToPlot

        metrics_of_interest = group[[plotkw['x'], plotkw['y']]]
        metric_corr_mat = metrics_of_interest.corr()
        metric_corr = metric_corr_mat.stack()
        metric_corr.name = 'corr'
        stack_idx = metric_corr.index
        valid_idxs = [(a, b) for (a, b) in ub.unique(map(tuple, map(sorted, stack_idx.to_list()))) if a != b]
        if valid_idxs:
            metric_corr = metric_corr.loc[valid_idxs]
            # corr_lbl = 'corr({},{})={:0.4f}'.format(*metric_corr.index[0], metric_corr.iloc[0])
            corr_lbl = 'corr={:0.4f}'.format(metric_corr.iloc[0])
        else:
            corr_lbl = ''
        data = group

        if huevar not in {'auto', 'random'}:
            allow_magic_huevar = True
            if allow_magic_huevar:
                if len(data[huevar].unique()) <= 1:
                    huevar = 'random'

        if huevar == 'random':
            import random
            huevar = random.choice(plotter.analysis.statistics)['param_name']
            # huevar = plotter.analysis.statistics[-1]['param_name']
            # huevar.replace('pxl', 'pred')
            plotkw['hue'] = huevar

        if huevar == 'auto':
            huevar = plotter.analysis.statistics[-1]['param_name']
            # huevar.replace('pxl', 'pred')
            plotkw['hue'] = huevar

        def make_fig(fnum, legend=True):
            fig = kwplot.figure(fnum=fnum, doclf=True)
            ax = fig.gca()
            n = len(data)
            ax = humanized_scatterplot(plotter.human_mapping, data=data, ax=ax,
                                       legend=legend, **plotkw)
            nice_type = plotter.human_mapping.get(type, type)
            ax.set_title(f'{nice_type} - {test_dset}\n{corr_lbl}, n={n}')

        prefix = f'{test_dset}_{type}_{huevar}'
        fnum = plot_name + prefix
        dpath = plotter.dpath
        run_make_fig(make_fig, fnum, dpath, plotter.human_mapping, plot_name, prefix)

    def plot_pixel_ap_verus_iarpa(plotter, code_type, group, huevar='sensorchan'):
        import kwplot
        plot_name = 'pxl_vs_iarpa'

        test_dset, type = code_type
        plotkw = ub.udict({
            'x': plotter.metric_luts['pxl'][type],
            'y': plotter.metric_luts['trk'][type],
            'hue': huevar,
            **plotter.common_plotkw,
        })

        missing = set((plotkw & {'x', 'y'}).values()) - set(group.columns)
        if missing:
            raise UnableToPlot(f'Cannot plot {plot_name} for {code_type} missing={missing}')

        if plotkw['x'] not in group.columns or plotkw['y'] not in group.columns:
            raise UnableToPlot

        metrics_of_interest = group[[plotkw['x'], plotkw['y']]]
        metric_corr_mat = metrics_of_interest.corr()
        metric_corr = metric_corr_mat.stack()
        metric_corr.name = 'corr'
        stack_idx = metric_corr.index
        valid_idxs = [(a, b) for (a, b) in ub.unique(map(tuple, map(sorted, stack_idx.to_list()))) if a != b]
        if valid_idxs:
            metric_corr = metric_corr.loc[valid_idxs]
            # corr_lbl = 'corr({},{})={:0.4f}'.format(*metric_corr.index[0], metric_corr.iloc[0])
            corr_lbl = 'corr={:0.4f}'.format(metric_corr.iloc[0])
        else:
            corr_lbl = ''
        data = group

        if huevar not in {'auto', 'random'}:
            allow_magic_huevar = True
            if allow_magic_huevar:
                if len(data[huevar].unique()) <= 1:
                    huevar = 'random'

        if huevar == 'random':
            import random
            huevar = random.choice(plotter.analysis.statistics)['param_name']
            # huevar = plotter.analysis.statistics[-1]['param_name']
            # huevar.replace('pxl', 'pred')
            plotkw['hue'] = huevar

        if huevar == 'auto':
            huevar = plotter.analysis.statistics[-1]['param_name']
            # huevar.replace('pxl', 'pred')
            plotkw['hue'] = huevar

        def make_fig(fnum, legend=True):
            fig = kwplot.figure(fnum=fnum, doclf=True)
            ax = fig.gca()
            n = len(data)
            ax = humanized_scatterplot(plotter.human_mapping, data=data, ax=ax,
                                       legend=legend, **plotkw)
            nice_type = plotter.human_mapping.get(type, type)
            ax.set_title(f'Pixelwise Vs IARPA metrics - {nice_type} - {test_dset}\n{corr_lbl}, n={n}')

        prefix = f'{test_dset}_{type}_{huevar}'
        fnum = plot_name + prefix
        dpath = plotter.dpath
        run_make_fig(make_fig, fnum, dpath, plotter.human_mapping, plot_name, prefix)

    def plot_pixel_ap_verus_auc(plotter, code_type, group, huevar='sensorchan'):
        import kwplot
        plot_name = 'pxl_vs_auc'
        # group = group[~group['sensorchan'].isnull()]
        # group['has_teamfeat'] = group['sensorchan'].apply(lambda x: (('depth' in x) or ('invariants' in x) or ('matseg' in x) or ('land' in x)))

        test_dset, type = code_type
        if type == 'eval_act+pxl':
            plotkw = ub.udict({
                'x': plotter.metric_luts['pxl'][type],
                'y': 'coi_mAUC',
                'hue': huevar,
                **plotter.common_plotkw,
            })
        elif type == 'eval_trk+pxl':
            plotkw = ub.udict({
                'x': plotter.metric_luts['pxl'][type],
                'y': 'salient_AUC',
                'hue': huevar,
                **plotter.common_plotkw,
            })
        else:
            raise KeyError(type)
        missing = set((plotkw & {'x', 'y'}).values()) - set(group.columns)
        if missing:
            raise UnableToPlot(f'Cannot plot plot_pixel_ap_verus_auc for {code_type} missing={missing}')

        metric_corr_mat = group[[plotkw['x'], plotkw['y']]].corr()
        metric_corr = metric_corr_mat.stack()
        metric_corr.name = 'corr'
        stack_idx = metric_corr.index
        valid_idxs = [(a, b) for (a, b) in ub.unique(map(tuple, map(sorted, stack_idx.to_list()))) if a != b]
        metric_corr = metric_corr.loc[valid_idxs]
        # corr_lbl = 'corr({},{})={:0.4f}'.format(*metric_corr.index[0], metric_corr.iloc[0])
        corr_lbl = 'corr={:0.4f}'.format(metric_corr.iloc[0])

        def make_fig(fnum, legend=True):
            fig = kwplot.figure(fnum=fnum)
            ax = fig.gca()
            ax = humanized_scatterplot(plotter.human_mapping, data=group,
                                       ax=ax, legend=legend, **plotkw)
            nice_type = plotter.human_mapping.get(type, type)
            ax.set_title(f'Pixelwise metrics - {nice_type} - {test_dset}\n{corr_lbl}')
            fig.set_size_inches(16.85, 8.82)

        prefix = f'{test_dset}_{type}_{huevar}'
        fnum = 'plot_pixel_ap_verus_auc' + prefix
        run_make_fig(make_fig, fnum, plotter.dpath, plotter.human_mapping, plot_name, prefix)

    def plot_param_analysis(plotter, code_type, group, metrics, params_of_interest=None):
        """
        metrics = ['salient_AP']
        metrics = ['BAS_F1']
        """
        test_dset, type = code_type
        metrics = metrics if ub.iterable(metrics) else [metrics]
        metrics_key = '_'.join(metrics)
        plot_name = 'violin_' + metrics_key
        prefix = f'{test_dset}_{type}_'

        metrics_key = '_'.join(metrics)
        # pred_param_df = pd.DataFrame(group['pred_params'].tolist(), index=group.index)
        # track_param_df = pd.DataFrame(group['track_params'].tolist(), index=group.index)
        # fit_param_df = pd.DataFrame(group['fit_params'].tolist(), index=group.index)
        # varied_params = ub.varied_values(pred_param_df.to_dict('records'), min_variations=1)
        # blocklist = ub.oset(['step', 'epoch', 'pred_in_dataset_name'])
        # pred_param_df = pred_param_df.drop(blocklist & pred_param_df.columns, axis=1)

        # main_cols = ub.oset(['expt', 'model', 'step', 'test_dset'] + metrics)
        expanded = group

        if params_of_interest is None:
            params_of_interest = [
                'pred_use_cloudmask',
                'pred_resample_invalid_frames',
                'pred_input_space_scale',
                'pred_window_space_scale',
                'trk_thresh',
            ]

        x = 'act.pxl.properties.model_name'

        additional_needed_legends = []

        def make_make_fig(expanded, param_name):
            def make_fig(fnum, legend=True):
                fig = kwplot.figure(fnum=fnum, doclf=True)
                ax = fig.gca()
                n = len(expanded)

                assert len(metrics) == 1

                # sns.violinplot(data=expanded, x=param_name, y=metrics[0], hue='expt')
                # sns.violinplot(data=expanded, x='expt', y=metrics[0], hue=param_name, split=True)
                # sns.boxplot(data=expanded, x='expt', y=metrics[0], hue=param_name, notch=True)

                if param_name not in expanded.columns:
                    raise UnableToPlot

                sns.boxplot(data=expanded, x=x, y=metrics[0], hue=param_name,
                            medianprops={"color": "coral"})
                if not legend:
                    ax.get_legend().remove()
                humanize_axes(ax, plotter.human_mapping)

                # Relabel if it's too big to fit.
                relabels = {}
                needs_relabel = False
                from itertools import count
                counter = count(1)

                xtick_labels = list(ax.get_xticklabels())
                n_xticks = len(xtick_labels)
                if n_xticks == 1:
                    len_thresh = 60
                else:
                    len_thresh = 10

                for label in xtick_labels:
                    text = label.get_text()
                    if len(text) > len_thresh:
                        needs_relabel = True
                        relabels[text] = str(next(counter))

                if needs_relabel:
                    additional = {
                        'relabels': relabels,
                        'param_name': param_name,
                    }
                    additional_needed_legends.append(additional)
                    new_labels = []
                    for label in ax.get_xticklabels():
                        text = label.get_text()
                        if text in relabels:
                            label.set_text(relabels[text])
                        new_labels.append(label)
                    ax.set_xticklabels(new_labels)

                # for label in ax.get_xticklabels():
                #     label.set_rotation(90)

                nice_type = plotter.human_mapping.get(type, type)
                nice_param_name = plotter.human_mapping.get(param_name, param_name)
                ax.set_title(f'Varied {nice_param_name} - {nice_type}\n{test_dset}\nn={n}')
                fig.set_size_inches(12.85, 8.82)
                fig.tight_layout()
            return make_fig

        import kwplot
        import seaborn as sns
        for param_name in params_of_interest:
            try:
                expanded[[param_name] + metrics]
            except KeyError:
                continue
            make_fig = make_make_fig(expanded, param_name)
            prefix = f'{test_dset}_{type}_'
            fnum = plot_name + param_name + prefix
            dpath = plotter.dpath
            try:
                run_make_fig(make_fig, fnum, dpath, plotter.human_mapping, plot_name + param_name, prefix)
            except Exception:
                print(f'Error checking {param_name}')
                ...
            # sns.violinplot(data=expanded, x=x, y=metrics[0], hue=param_name,
            #                medianprops={"color": "coral"})

        # In the case we needed to relabel an axis, we also need to write a
        # figure indicating what that relabel was.
        for additional in additional_needed_legends:
            relabels = additional['relabels']
            param_name = additional['param_name']
            df = pd.DataFrame([{'orig': k, 'relabel': v} for k, v in relabels.items()])
            df = df.set_index('orig')
            df2_style = humanize_dataframe(df, title='Relabling')

            # Hack, write the new figure to the parts dir. We have to be sure
            # to reconstruct it correctly.

            plot_dpath_parts = (dpath / (plot_name + param_name + '_parts')).ensuredir()
            fpath = plot_dpath_parts / prefix + 'relabel.png'
            dfi_table(df2_style, fpath, fontsize=32, show=False)

    def plot_resource_versus_metric(plotter, code_type, group, huevar='sensorchan'):
        import kwplot
        plot_name = 'resource_vs_metric'

        resources_of_interest = [
            'trk.pxl.resource.total_hours',
            'trk.pxl.resource.co2_kg',
            'act.pxl.resource.total_hours',
            'act.pxl.resource.co2_kg',
            'act.poly.resource.total_hours',
            'trk.poly.resource.total_hours',
        ]

        metrics = [
            'trk.poly.metrics.sc_macro_f1',
        ]

        for resource_type in resources_of_interest:
            human_resource_type = plotter.human_mapping.get(resource_type, resource_type)

            for metric in metrics:
                human_metric_type = plotter.human_mapping.get(metric, metric)

                # group['pred_tta_time'] = group['pred_tta_time'].astype(str)
                # group['pred_tta_fliprot'] = group['pred_tta_fliprot'].astype(str)
                # group.loc[group['pred_tta_time'] == 'nan', 'pred_tta_time'] = '0.0'
                # group.loc[group['pred_tta_fliprot'] == 'nan', 'pred_tta_fliprot'] = '0.0'

                test_dset, type = code_type
                plotkw = ub.udict({
                    'x': resource_type,
                    'y': metric,
                    'hue': huevar,
                    **plotter.common_plotkw,
                    'style': 'hardware',
                })

                missing = set((plotkw & {'x', 'y'}).values()) - set(group.columns)
                if missing:
                    print(f'Cannot plot plot_resource_versus_metric for {code_type} missing={missing}')
                    continue

                def make_fig(fnum, legend=True):
                    fig = kwplot.figure(fnum=fnum)
                    ax = fig.gca()
                    ax = humanized_scatterplot(plotter.human_mapping, data=group, ax=ax, **plotkw)
                    nice_type = plotter.human_mapping.get(type, type)
                    ax.set_title(f'{human_resource_type} vs {human_metric_type} - {nice_type} - {test_dset}')
                    fig.set_size_inches(np.array([6.4, 4.8]) * 1.4)

                prefix = f'{test_dset}_{type}_'
                fnum = 'plot_resource_versus_metric_' + resource_type + prefix
                run_make_fig(make_fig, fnum, plotter.dpath, plotter.human_mapping, plot_name, prefix)


def humanize_dataframe(df, col_formats=None, human_labels=None, index_format=None,
                       title=None):
    import humanize
    df2 = df.copy()
    if col_formats is not None:
        for col, fmt in col_formats.items():
            if fmt == 'intcomma':
                df2[col] = df[col].apply(humanize.intcomma)
            if fmt == 'concice_si_display':
                from kwcoco.metrics.drawing import concice_si_display
                for row in df2.index:
                    val = df2.loc[row, col]
                    if isinstance(val, float):
                        val = concice_si_display(val)
                        df2.loc[row, col] = val
                df2[col] = df[col].apply(humanize.intcomma)
            if callable(fmt):
                df2[col] = df[col].apply(fmt)
    if human_labels:
        df2 = df2.rename(human_labels, axis=1)

    indexes = [df2.index, df2.columns]
    if human_labels:
        for index in indexes:
            if index.name is not None:
                index.name = human_labels.get(index.name, index.name)
            if index.names:
                index.names = [human_labels.get(n, n) for n in index.names]

    if index_format == 'capcase':
        def capcase(x):
            if '_' in x or x.islower():
                return ' '.join([w.capitalize() for w in x.split('_')])
            return x
        df2.index.values[:] = [human_labels.get(x, x) for x in df2.index.values]
        df2.index.values[:] = list(map(capcase, df2.index.values))
        # human_df = human_df.applymap(lambda x: str(x) if isinstance(x, int) else '{:0.2f}'.format(x))
        pass

    df2_style = df2.style
    if title:
        df2_style = df2_style.set_caption(title)
    return df2_style


def humanize_legend(legend, human_mapping):
    leg_title = legend.get_title()
    old_text = leg_title.get_text()
    new_text = human_mapping.get(old_text, old_text)
    leg_title.set_text(new_text)
    for leg_lbl in legend.texts:
        old_text = leg_lbl.get_text()
        new_text = human_mapping.get(old_text, old_text)
        leg_lbl.set_text(new_text)


def humanize_axes(ax, human_mapping):
    xkey = ax.get_xlabel()
    ykey = ax.get_ylabel()
    ax.set_xlabel(human_mapping.get(xkey, xkey))
    ax.set_ylabel(human_mapping.get(ykey, ykey))
    legend = ax.get_legend()
    if legend is not None:
        humanize_legend(legend, human_mapping)


def humanized_scatterplot(human_mapping, data, ax, plot_type='scatter', mesh=None, connect=None, star=None, starkw=None, **plotkw):
    """
    Example:
        import pandas as pd
        human_mapping = {}
        ax = None
        plotkw = {'x': 'x', 'y': 'y', 'hue': 'group'}
        n = 100
        data = pd.DataFrame({
             'x': np.random.rand(n),
             'y': np.random.rand(n),
             'group': (np.random.rand(n) * 5).astype(int),
             'star': (np.random.rand(n) > 0.9).astype(int),
        })
        mesh = 'group'
        import kwplot
        kwplot.autompl()
        kwplot.figure(fnum=32)
        humanized_scatterplot(human_mapping, data, ax, mesh, **plotkw)
    """
    import seaborn as sns
    import kwplot
    import kwimage
    plt = kwplot.autoplt()
    xkey = plotkw['x']
    ykey = plotkw['y']

    if ax is None:
        import kwplot
        plt = kwplot.autoplt()
        ax = plt.gca()

    if star is not None:
        if star in data:
            _starkw = ub.dict_isect(plotkw, {'s'})
            _starkw = {
                's': _starkw.get('s', 10) + 280,
                'color': 'orange',
            }
            if starkw is not None:
                _starkw.update(starkw)
            flags = data[star].apply(bool)
            star_data = data[flags]
            star_x = star_data[xkey]
            star_y = star_data[ykey]
            ax.scatter(star_x, star_y, marker='*', **_starkw)

    if plot_type == 'scatter':
        ax = sns.scatterplot(data=data, ax=ax, **plotkw)
    else:
        ax = sns.lineplot(data=data, ax=ax, **plotkw)

    humanize_axes(ax, human_mapping)

    if connect:
        import scipy
        import scipy.spatial
        groups = data.groupby(connect)
        colors = kwimage.Color.distinct(len(groups))
        i = 0
        for gkey, subgroup in groups:
            if 'step' in subgroup.columns:
                subgroup = subgroup.sort_values('epoch')
            points = subgroup[[xkey, ykey]].values
            ax = plt.gca()
            ax.plot(points[:, 0], points[:, 1], '--', alpha=0.3, color='gray')
            i += 1

    if mesh:
        import scipy
        import scipy.spatial
        mesh_groups = data.groupby(mesh)
        colors = kwimage.Color.distinct(len(mesh_groups))
        i = 0
        for gkey, subgroup in mesh_groups:
            points = subgroup[[xkey, ykey]].values
            did_plot = 0
            if 1:
                if len(points) > 3:
                    try:
                        tri = scipy.spatial.Delaunay(points)
                        if 0:
                            # MSE
                            # todo: cut off non-mse edges, or order based on epoch?
                            import networkx as nx
                            g = nx.Graph()
                            g.add_edges_from(tri.simplices[:, 0:2])
                            g.add_edges_from(tri.simplices[:, 1:3])
                            g.add_edges_from(tri.simplices[:, [2, 0]])
                            mse_edges = list(nx.minimum_spanning_tree(g).edges)
                            segments = points[mse_edges, :]
                            pts1 = segments[:, 0, :]
                            pts2 = segments[:, 1, :]
                            kwplot.draw_line_segments(pts1, pts2, color=colors[i])
                        else:
                            plt.triplot(points[:, 0], points[:, 1], tri.simplices, alpha=0.2)
                    except Exception:
                        pass
                    else:
                        did_plot = 1
            if not did_plot:
                # Just trace the points in whatever order
                ax = plt.gca()
                ax.plot(points[:, 0], points[:, 1], alpha=0.2, color='gray')
            i += 1
    return ax


def describe_varied(merged_df, dpath, human_mapping=None):
    # import pprint
    expt_group = dict(list(merged_df.groupby(['dataset_code', 'type'])))
    fnum = 40

    human_mapping.update({
        'time_steps': 'Time Steps (frames)',
        # 'chip_size': 'Chip Size (pxls)',  # TODO chip_dims
        'time_span': 'Time Span',
        'time_sampling': 'Temporal Sampling Method',
        'num_unique': 'Num Unique',
        'top_val': 'Most Frequent Value',
        'top_freq': 'Top Freq',
        'param': 'Param Name',
        'init': 'Network Initialization',
        'pred_tta_time': 'Temporal Test Time Augmentation',
        'pred_tta_time': 'Temporal Test Time Augmentation',
    })

    ignore_params = {
        'bad_channels',
        'true_multimodal',
        'modulate_class_weights',
        'channels',
    }

    varied_dpath = (dpath / 'varied_params').ensuredir()

    def varied_param_table(fnum, param_keys, dataset_code, param_type):
        have_params = list(ub.oset(param_keys) & set(group.columns))
        if len(have_params) == 0:
            return
        part = group[have_params].fillna('null')
        # varied_series = part.nunique()
        # varied_series = varied_series[varied_series > 1]
        # print(varied_series)
        # human_df = varied_df.rename(human_mapping, axis=0)

        param_to_row = ub.ddict(dict)
        param_to_hist = {k: part[k].value_counts() for k in have_params}
        rows = []
        for param, hist in param_to_hist.items():
            if param in ignore_params:
                continue
            row = param_to_row[param]
            row['param'] = param
            row['num_unique'] = len(hist)
            row['top_val'] = hist.idxmax()
            # row['top_freq'] = hist.max()
            if row['num_unique'] > 1:
                rows.append(row)
        if len(rows) == 0:
            return None
        param_summary = pd.DataFrame(rows).set_index('param', drop=True)

        col_formats = {
            'num_unique': int,
            'top_val': 'concice_si_display',
        }
        index_format = 'capcase'
        title = f'Varied {param_type} Parameters: {dataset_code}'
        df = param_summary
        df2_style = humanize_dataframe(df, col_formats,
                                       index_format=index_format,
                                       human_labels=human_mapping,
                                       title=title)
        fname = 'varied_' + dataset_code + '_' + param_type + '_' + ub.hash_data([fnum, code_type])[0:16] + '.png'
        fpath = varied_dpath / fname
        dfi_table(df2_style, fpath, fontsize=12, show=False)

    for code_type, group in expt_group.items():
        print(f'code_type={code_type}')
        fnum += 1
        dataset_code, type = code_type

        print('Varied fit params')
        print('type = {}'.format(ub.urepr(type, nl=1)))
        print('dataset_code = {}'.format(ub.urepr(dataset_code, nl=1)))
        param_keys = heuristics.fit_param_keys
        varied_param_table(fnum, param_keys, dataset_code, param_type='Fit')
        # print(ub.highlight_code(pprint.pformat(dict(varied), width=80)))
        # print('varied = {}'.format(ub.urepr(varied, nl=2)))

        print('Varied pred params')
        print('type = {}'.format(ub.urepr(type, nl=1)))
        print('dataset_code = {}'.format(ub.urepr(dataset_code, nl=1)))
        param_keys = pred_param_keys
        fnum += 1
        varied_param_table(fnum, param_keys, dataset_code, param_type='Predict')
        # part = group[pred_param_keys].fillna('null')
        # rows = part.to_dict('records')
        # varied = ub.varied_values(rows, 0)
        # print(ub.highlight_code(pprint.pformat(dict(varied), width=80)))

        print('Varied track params')
        print('type = {}'.format(ub.urepr(type, nl=1)))
        print('dataset_code = {}'.format(ub.urepr(dataset_code, nl=1)))
        param_keys = trk_param_keys
        fnum += 1
        varied_param_table(fnum, param_keys, dataset_code, param_type='BAS Tracking')
        # part = group[trk_param_keys].fillna('null')
        # rows = part.to_dict('records')
        # varied = ub.varied_values(rows, 0)
        # print(ub.highlight_code(pprint.pformat(dict(varied), width=80)))

        print('Varied activity params')
        print('type = {}'.format(ub.urepr(type, nl=1)))
        print('dataset_code = {}'.format(ub.urepr(dataset_code, nl=1)))
        param_keys = act_param_keys
        fnum += 1
        varied_param_table(fnum, param_keys, dataset_code, param_type='SC Classification')
        # part = group[act_param_keys].fillna('null')
        # rows = part.to_dict('records')
        # varied = ub.varied_values(rows, 0)
        # print(ub.highlight_code(pprint.pformat(dict(varied), width=80)))


memo_kwcoco_load = ub.memoize(kwcoco.CocoDataset)


def dataset_summary_tables(dpath):
    import watch

    expt_dvc_dpath = watch.find_smart_dvc_dpath()
    rows = []
    DSET_CODE_TO_TASK = {
        # 'Aligned-Drop3-TA1-2022-03-10': 'bas',
        # 'Aligned-Drop3-L1': 'bas',
        # 'Cropped-Drop3-TA1-2022-03-10': 'sc',
        'Aligned-Drop4-2022-08-08-TA1-S2-L8-ACC': 'bas',
        'Aligned-Drop4-2022-08-08-TA1-S2-WV-PD-ACC': 'sc',
    }
    for bundle_name in DSET_CODE_TO_TASK.keys():
        task = DSET_CODE_TO_TASK[bundle_name]
        gsd = DSET_CODE_TO_GSD.get(bundle_name, None)
        bundle_dpath = expt_dvc_dpath / bundle_name
        train_fpath = bundle_dpath / 'data_train.kwcoco.json'
        vali_fpath = bundle_dpath / 'data_vali.kwcoco.json'

        if not train_fpath.exists():
            raise Exception
            watch.utils.simple_dvc.SimpleDVC().pull(list(bundle_dpath.glob('splits*.dvc')))
            zip_fpaths = list(bundle_dpath.glob('splits.zip'))
            for p in zip_fpaths:
                ub.cmd(f'7z x {p}', verbose=3, check=1, cwd=bundle_dpath)
            pass

        print(f'read train_fpath={train_fpath}')
        train_dset = memo_kwcoco_load(train_fpath)
        print(f'read vali_fpath={vali_fpath}')
        vali_dset = memo_kwcoco_load(vali_fpath)

        if gsd is None:
            gsd = set(vid['target_gsd'] for vid in train_dset.videos().objs)

        if ub.iterable(gsd) and len(gsd) == 1:
            gsd = ub.peek(gsd)

        type_to_dset = {'train': train_dset, 'vali': vali_dset}
        for split, dset in type_to_dset.items():
            unique_regions = {'_'.join(n.split('_')[0:2]) for n in dset.videos().get('name')}
            unique_sites = set(dset.annots().get('track_id'))
            num_tracks = len(unique_sites)
            row = {
                'dataset': bundle_name,
                'task': task,
                'split': split,
                'gsd': gsd,
                'num_regions': len(unique_regions),
                'num_sites': num_tracks,
                'num_videos': dset.n_videos,
                'num_images': dset.n_images,
                'num_annots': dset.n_annots,
            }
            rows.append(row)

    human_labels = {
        'dataset': 'Dataset Codename',
        'task': 'Task',
        'split': 'Split',
        'gsd': 'GSD',
        'num_sites': 'Num Sites',
        'num_videos': 'Num Videos',
        'num_regions': 'Num Regions',
        'num_images': 'Num Images',
        'num_annots': 'Num Annots',
    }

    col_formats = {
        'gsd': 'intcomma',
        'num_sites': 'intcomma',
        'num_videos': 'intcomma',
        'num_images': 'intcomma',
        'num_annots': 'intcomma',
    }
    df = pd.DataFrame(rows)
    df['task'] = df['task'].apply(str.upper)
    df = df.set_index(['dataset', 'task', 'split'])
    import rich
    rich.print(df.to_string())
    title = 'Dataset Summary'
    df2_style = humanize_dataframe(df, col_formats, human_labels=human_labels,
                                   title=title)
    fpath = dpath / 'dataset_summary.png'
    dfi_table(df2_style, fpath, fontsize=32, show='eog')


def initial_summary(reporter, dpath=None):
    table = reporter.orig_merged_df.copy()
    # Alternate way to compute
    co2_rows = []
    kwh_rows = []
    hour_rows = []

    # type_to_rows = dict(list(reporter.orig_merged_df.groupby('type')))
    # type_to_resource = {}
    # pairs = [
    #     ('eval_trk_poly_fpath', 'trk.poly.resource.total_hours'),
    #     ('eval_act_poly_fpath', 'trk.poly.resource.total_hours'),
    # ]
    # if k1 in type_to_rows:
    #     type_to_rows[k1][k2].sum()

    reporter.orig_merged_df['trk.pxl.resource.co2_kg']
    reporter.orig_merged_df['trk.pxl.resource.total_hours']
    reporter.orig_merged_df['act.pxl.resource.co2_kg']
    reporter.orig_merged_df['act.pxl.resource.total_hours']

    # for row in reporter.merg
    for row in reporter.big_rows:
        if row['type'] == 'eval_pxl':
            co2_rows += [row['info']['param_types']['resource'].get('co2_kg', np.nan)]
            kwh_rows += [row['info']['param_types']['resource'].get('kwh', np.nan)]
            hour_rows += [row['info']['param_types']['resource'].get('total_hours', np.nan)]

    def mean_imputer(arr):
        arr = np.array([np.nan if a is None else a for a in arr])
        val = np.nanmean(arr)
        arr[np.isnan(arr)] = val
        return arr

    co2_rows = mean_imputer(np.array(co2_rows))
    hour_rows = mean_imputer(np.array(hour_rows))
    kwh_rows = mean_imputer(np.array(kwh_rows))

    total_co2 = co2_rows.sum()
    total_hours = hour_rows.sum()
    total_kwh = kwh_rows.sum()

    num_expt_models = table[['expt', 'model']].nunique()
    num_eval_types = table['type'].value_counts()

    summary_df = pd.concat([num_expt_models, num_eval_types]).to_frame().T
    order = ub.oset(['expt', 'model', 'eval_pxl', 'eval_trk', 'eval_act']) & summary_df.columns
    summary_df = summary_df[order]
    summary_df['co2_kg'] = [total_co2]
    summary_df['kwh'] = [total_kwh]
    summary_df['hours'] = [total_hours]
    summary_df['index'] = ['Total']
    summary_df = summary_df.set_index('index', drop=True)
    summary_df.index.name = None
    humanize = {
        'expt': 'Num Training Runs (Experiments)',
        'model': 'Num Checkpoints Selected (Models)',
        'eval_pxl': 'Num Pixel Evaluations',
        'eval_trk': 'Num IARPA BAS Evaluations',
        'eval_act': 'Num IARPA SC Evaluations',
        'co2_kg': 'Total Carbon Cost (kg)',
        'kwh': 'Energy (kW/H)',
        'hours': 'Total Compute Time (hours)',
    }
    human_df = summary_df.rename(humanize, axis=1)
    human_df = human_df.applymap(lambda x: str(x) if isinstance(x, int) else '{:0.2f}'.format(x))
    table_style = human_df.T.style
    table_style = table_style.set_caption('Experimental Summary')

    import rich
    rich.print(human_df)

    fpath = dpath / 'big_picture_experiment_summary.png'
    dfi_table(table_style, fpath, fontsize=24, show='eog')


def dfi_table(table_style, fpath, fontsize=12, fnum=None, show='eog'):
    import kwimage
    import kwplot
    import dataframe_image as dfi
    dfi_converter = "chrome"  # matplotlib
    dfi.export(
        table_style,
        str(fpath),
        table_conversion=dfi_converter,
        fontsize=fontsize,
        max_rows=-1,
    )
    if show == 'imshow':
        imdata = kwimage.imread(fpath)
        kwplot.imshow(imdata, fnum=fnum)
    elif show == 'eog':
        import xdev
        xdev.startfile(fpath)


def run_make_fig(make_fig, fnum, dpath, human_mapping, plot_name, prefix):
    """
    Runs a function that plots a figure with and without a legend.
    Also saves the legend to its own file.
    """
    import kwplot
    plt = kwplot.autoplt()

    plot_dpath_main = (dpath / plot_name).ensuredir()
    plot_dpath_parts = (dpath / (plot_name + '_parts')).ensuredir()

    print(f'fnum={fnum}')
    make_fig(str(fnum) + '_legend', legend=True)
    fig = plt.gcf()
    # fname = f'{prefix}{plot_name}.png'
    fname = f'{fnum}.png'
    fpath = plot_dpath_main / fname
    fig.set_size_inches(np.array([6.4, 4.8]) * 1.4)
    fig.tight_layout()
    fig.savefig(fpath)

    SAVE_PARTS = 1
    if SAVE_PARTS:
        ax = fig.gca()
        orig_legend = ax.get_legend()
        orig_legend_title = orig_legend.get_title().get_text()
        legend_handles = ax.get_legend_handles_labels()

        make_fig(str(fnum) + '_nolegend', legend=False)
        fig_nolegend = plt.gcf()
        fname = f'{prefix}{plot_name}_nolegend.png'
        fpath = plot_dpath_parts / fname
        fig_nolegend.set_size_inches(np.array([6.4, 4.8]) * 1.4)
        fig_nolegend.tight_layout()
        fig_nolegend.savefig(fpath)

        fig_onlylegend = kwplot.figure(
            fnum=str(fnum) + '_onlylegend', doclf=1)
        ax2 = fig_onlylegend.gca()
        ax2.axis('off')
        new_legend = ax2.legend(*legend_handles, title=orig_legend_title,
                                loc='lower center')
        humanize_legend(new_legend, human_mapping)
        fname = f'{prefix}{plot_name}_onlylegend.png'
        fpath = plot_dpath_parts / fname
        try:
            new_extent = new_legend.get_window_extent()
            inv_scale = fig_onlylegend.dpi_scale_trans.inverted()
            bbox = new_extent.transformed(inv_scale)
            newkw = {'bbox_inches': bbox}
        except Exception:
            newkw = {'bbox_inches': None}
        fig_onlylegend.tight_layout()
        fig_onlylegend.savefig(fpath, **newkw)
        kwplot.close_figures([fig_onlylegend, fig_nolegend])
        from watch.utils import util_kwplot
        util_kwplot.cropwhite_ondisk(fpath)



import copy
import numpy as np
import ubelt as ub
from collections import OrderedDict
import decimal
import fractions
import pathlib


def debug_json_unserializable(data, msg=''):
    """
    Raises an exception if the data is not serializable and prints information
    about it.
    """
    from kwcoco.util import util_json
    import ubelt as ub
    unserializable = list(util_json.find_json_unserializable(data))
    if unserializable:
        raise Exception(msg + ub.urepr(unserializable))


def ensure_json_serializable(dict_, normalize_containers=False, verbose=0):
    """
    Attempt to convert common types (e.g. numpy) into something json complient

    Convert numpy and tuples into lists

    Args:
        normalize_containers (bool):
            if True, normalizes dict containers to be standard python
            structures. Defaults to False.

    Example:
        >>> from watch.utils.util_json import *  # NOQA
        >>> from kwcoco.util import util_json
        >>> data = ub.ddict(lambda: int)
        >>> data['foo'] = ub.ddict(lambda: int)
        >>> data['bar'] = np.array([1, 2, 3])
        >>> data['foo']['a'] = 1
        >>> data['foo']['b'] = (1, np.array([1, 2, 3]), {3: np.int32(3), 4: np.float16(1.0)})
        >>> dict_ = data
        >>> print(ub.urepr(data, nl=-1))
        >>> assert list(util_json.find_json_unserializable(data))
        >>> result = ensure_json_serializable(data, normalize_containers=True)
        >>> print(ub.urepr(result, nl=-1))
        >>> assert not list(util_json.find_json_unserializable(result))
        >>> assert type(result) is dict
    """
    dict_ = copy.deepcopy(dict_)

    def _norm_container(c):
        if isinstance(c, dict):
            # Cast to a normal dictionary
            if isinstance(c, OrderedDict):
                if type(c) is not OrderedDict:
                    c = OrderedDict(c)
            else:
                if type(c) is not dict:
                    c = dict(c)
        return c

    walker = ub.IndexableWalker(dict_)
    for prefix, value in walker:
        if isinstance(value, tuple):
            new_value = list(value)
            walker[prefix] = new_value
        elif isinstance(value, np.ndarray):
            new_value = value.tolist()
            walker[prefix] = new_value
        elif isinstance(value, (np.integer)):
            new_value = int(value)
            walker[prefix] = new_value
        elif isinstance(value, (np.floating)):
            new_value = float(value)
            walker[prefix] = new_value
        elif isinstance(value, (np.complexfloating)):
            new_value = complex(value)
            walker[prefix] = new_value
        elif isinstance(value, decimal.Decimal):
            new_value = float(value)
            walker[prefix] = new_value
        elif isinstance(value, fractions.Fraction):
            new_value = float(value)
            walker[prefix] = new_value
        elif isinstance(value, pathlib.Path):
            new_value = str(value)
            walker[prefix] = new_value
        elif hasattr(value, '__json__'):
            new_value = value.__json__()
            walker[prefix] = new_value
        elif normalize_containers:
            if isinstance(value, dict):
                new_value = _norm_container(value)
                walker[prefix] = new_value

    if normalize_containers:
        # normalize the outer layer
        dict_ = _norm_container(dict_)
    return dict_



"""
containing some widely-used pre-processing method for change detection
"""
import numpy as np


def norm_img(img, channel_first=True):
    if channel_first:
        channel, img_height, img_width = img.shape
        img = np.reshape(img, (channel, img_height * img_width))  # (channel, height * width)
        max_value = np.max(img, axis=1, keepdims=True)  # (channel, 1)
        min_value = np.min(img, axis=1, keepdims=True)  # (channel, 1)
        diff_value = max_value - min_value
        nm_img = (img - min_value) / diff_value
        nm_img = np.reshape(nm_img, (channel, img_height, img_width))
    else:
        img_height, img_width, channel = img.shape
        img = np.reshape(img, (img_height * img_width, channel))  # (channel, height * width)
        max_value = np.max(img, axis=0, keepdims=True)  # (channel, 1)
        min_value = np.min(img, axis=0, keepdims=True)  # (channel, 1)
        diff_value = max_value - min_value
        nm_img = (img - min_value) / diff_value
        nm_img = np.reshape(nm_img, (img_height, img_width, channel))
    return nm_img


def stad_img(img, channel_first=True, get_para=False):
    """
    normalization image
    :param channel_first:
    :param img: (C, H, W)
    :return:
        norm_img: (C, H, W)
    """
    if channel_first:
        channel, img_height, img_width = img.shape
        img = np.reshape(img, (channel, img_height * img_width))  # (channel, height * width)
        mean = np.mean(img, axis=1, keepdims=True)  # (channel, 1)
        center = img - mean  # (channel, height * width)
        var = np.sum(np.power(center, 2), axis=1, keepdims=True) / (img_height * img_width)  # (channel, 1)
        std = np.sqrt(var)  # (channel, 1)
        nm_img = center / std  # (channel, height * width)
        nm_img = np.reshape(nm_img, (channel, img_height, img_width))
    else:
        img_height, img_width, channel = img.shape
        img = np.reshape(img, (img_height * img_width, channel))  # (height * width, channel)
        mean = np.mean(img, axis=0, keepdims=True)  # (1, channel)
        center = img - mean  # (height * width, channel)
        var = np.sum(np.power(center, 2), axis=0, keepdims=True) / (img_height * img_width)  # (1, channel)
        std = np.sqrt(var)  # (channel, 1)
        nm_img = center / std  # (channel, height * width)
        nm_img = np.reshape(nm_img, (img_height, img_width, channel))
    print('mean is ', mean)
    print('std is ', std)
    if get_para:
        return nm_img, mean, std
    else:
        return nm_img


def CVA(img_X, img_Y, stad=False):
    # CVA has not affinity transformation consistency, so it is necessary to normalize multi-temporal images to
    # eliminate the radiometric inconsistency between them
    if stad:
        img_X = stad_img(img_X)
        img_Y = stad_img(img_Y)
    img_diff = img_X - img_Y
    L2_norm = np.sqrt(np.sum(np.square(img_diff), axis=0))
    return L2_norm


def main():
    from osgeo import gdal
    import time
    import imageio

    data_set_X = gdal.Open('../../../Dataset/Landsat/Taizhou/2000TM')  # data set X
    data_set_Y = gdal.Open('../../../Dataset/Landsat/Taizhou/2003TM')  # data set Y

    img_width = data_set_X.RasterXSize  # image width
    img_height = data_set_X.RasterYSize  # image height

    img_X = data_set_X.ReadAsArray(0, 0, img_width, img_height)
    img_Y = data_set_Y.ReadAsArray(0, 0, img_width, img_height)

    channel, img_height, img_width = img_X.shape
    tic = time.time()
    L2_norm = CVA(img_X, img_Y)

    bcm = np.ones((img_height, img_width))

    raise NotImplementedError
    otsu = None
    thre = otsu(L2_norm.reshape(1, -1))
    bcm[L2_norm > thre] = 255
    bcm = np.reshape(bcm, (img_height, img_width))
    imageio.imwrite('CVA_Taizhou.png', bcm)
    toc = time.time()
    print(toc - tic)


if __name__ == '__main__':
    main()
